import pytest
from stograde.common.find_unmerged_branches_in_cwd import find_unmerged_branches_in_cwd
from stograde.common.run import run


def git(cmd, *args):
    return run(['git', cmd, *args])


def touch(file):
    return run(['touch', file])


@pytest.mark.skip(reason="this fails on travis-ci")
def test_find_unmerged_branches_in_cwd_1(tmpdir):
    with tmpdir.as_cwd():
        git('init')

        touch('file1')
        git('add', 'file1')
        git('commit', '-m', 'initial')

        git('checkout', '-b', 'branch')

        touch('file2')
        git('add', 'file2')
        git('commit', '-m', 'newcommit')

        git('checkout', 'master')

        print(git('--version'))

        assert find_unmerged_branches_in_cwd() == ['branch']


@pytest.mark.skip(reason="this fails on travis-ci")
def test_find_unmerged_branches_in_cwd_2(tmpdir):
    with tmpdir.as_cwd():
        git('init')

        touch('file1')
        git('add', 'file1')
        git('commit', '-m', 'initial')

        git('checkout', '-b', 'branch')

        touch('file2')
        git('add', 'file2')
        git('commit', '-m', 'newcommit')

        git('checkout', 'master')
        git('merge', 'branch')

        assert find_unmerged_branches_in_cwd() == []



from typing import List

from .run import run


# TODO: This is a bit slow. Look for a faster way to check.
def find_unmerged_branches_in_cwd() -> List[str]:
    """Check for unmerged branches in the current repository"""
    _, unmerged_branches, _ = run(['git', 'branch', '-a', '--no-merged', 'master'])
    return [s.strip()
            for s in unmerged_branches.split('\n')
            if s.strip()]



import os


def dirsize(path: str = '.') -> int:
    """Get the size of a folder (including descendants)"""
    total_size = 0
    for dirpath, _, filenames in os.walk(path):
        for f in filenames:
            if not f.startswith('.'):
                try:
                    dir_size = os.path.getsize(os.path.join(dirpath, f))
                except OSError:
                    dir_size = 0
                total_size += dir_size
    return total_size



# from http://stackoverflow.com/a/2158532/2347774
def flatten(lst):
    """Flatten a list"""
    for elem in lst:
        if isinstance(elem, list) and not isinstance(elem, str):
            yield from flatten(elem)
        else:
            yield elem



import logging
from typing import List


def load_stogradeignore() -> List[str]:
    try:
        with open('.stogradeignore', encoding='utf-8') as infile:
            ignored_specs = [line.strip() for line in infile.read().splitlines()]
            logging.debug("Ignored specs: {}".format(ignored_specs))
    except FileNotFoundError:
        logging.debug("No .stogradeignore file found")
        return []



from dataclasses import dataclass

from ..common.run_status import RunStatus


@dataclass
class CompileResult:
    """Result from compiling an assignment file"""
    command: str  # Command used to compile
    output: str     # Output from running the command
    status: RunStatus



from stograde.common import parse_commit_msg_for_assignments
from stograde.common import flatten
from natsort import natsorted


def parse_commits_for_assignments(commits):
    """Takes a list of commits and returns the affected assignments

    Input is a commit message; it returns a list of (hw|lab, ##) tuples.

    A commit looks like the following:

    {
      "id": "b6568db1bc1dcd7f8b4d5a946b0b91f9dacd7327",
      "message": "Update Catalan translation to e38cb41.",
      "timestamp": "2011-12-12T14:27:31+02:00",
      "url": "http://example.com/mike/diaspora/commit/b6568db1bc1dcd7f8b4d5a946b0b91f9dacd7327",
      "author": {
        "name": "Jordi Mallach",
        "email": "jordi@softcatala.org"
      },
      "added": ["CHANGELOG"],
      "modified": ["app/controller/application.rb"],
      "removed": []
    }

    Tested on these commit messages:
        - 'hw13 complete; part of hw14'
        - [more messages in test/assignment_parsing_test]
    """
    assignments = [parse_commit_msg_for_assignments(c['message']) for c in commits]
    return natsorted(set(flatten(assignments)))



from appdirs import AppDirs
from configparser import ConfigParser
from datetime import datetime, timedelta
import os

_dirs = AppDirs('stograde', 'StoDevX')


class Config:
    _filename = os.path.join(_dirs.user_config_dir, 'stograde.ini')
    _config = ConfigParser()

    def __init__(self):
        # load the config file
        loaded_files = self._config.read(self._filename)
        # create the config file if it doesn't exist
        if not loaded_files:
            self._setup()

    def _setup(self):
        # create the config dir
        os.makedirs(_dirs.user_config_dir, exist_ok=True)
        # initialize with default values
        self._config['general'] = {}
        self.set_last_update_check()
        # save the file b/c it didn't exist
        self.save_config()

    def get_last_update_check(self):
        value = self._config.get('general', 'last_update_check', fallback=str(datetime.now()))
        return datetime.strptime(value, '%Y-%m-%d %H:%M:%S.%f')

    def set_last_update_check(self):
        self._config.set('general', 'last_update_check', str(datetime.now()))
        self.save_config()

    def needs_update_check(self):
        return self.get_last_update_check() < datetime.now() - timedelta(minutes=15)

    def save_config(self):
        with open(self._filename, 'w', encoding='utf-8') as outfile:
            self._config.write(outfile)


conf = Config()



from dataclasses import dataclass


@dataclass
class FileOptions:
    compile_optional: bool = False
    hide_contents: bool = False
    optional: bool = False
    timeout: int = 4
    truncate_contents: int = 10000
    truncate_output: int = 10000
    web_file: bool = False

    def update(self, options: dict):
        self.compile_optional = options.get('optional_compile', self.compile_optional)
        self.hide_contents = options.get('hide_contents', self.hide_contents)
        self.optional = options.get('optional', self.optional)
        self.timeout = options.get('timeout', self.timeout)
        self.truncate_contents = options.get('truncate_contents', self.truncate_contents)
        self.truncate_output = options.get('truncate_output', self.truncate_output)
        self.web_file = options.get('web', self.web_file)
        return self



import html
import traceback


def format_assignment_html(recording, debug=False):
    """Given a single recording, format it into an HTML file.

    Each recording will only have one student.

    Returns a {content: str, student: str, type: str, assignment: str} dict.
    """

    try:
        files = format_files_list(recording.get('files', {}))
        warnings = format_warnings(recording.get('warnings', {}).items())
        header = format_header(recording, warnings)
        output = (header + files) + '\n\n'

    except Exception as err:
        if debug:
            raise err
        output = format_as_code(traceback.format_exc())

    return {
        'assignment': recording['spec'],
        'content': output,
        'student': recording['student'],
        'type': 'html',
    }


def format_files_list(files):
    return '\n\n'.join([format_file(name, info) for name, info in files.items()])


def format_warnings(warnings):
    formatted = [format_warning(warning, value) for warning, value in warnings]
    return [w for w in formatted if w]


def format_header(recording, warnings):
    header = '<h1>{spec}</h1>'.format_map(recording)

    if warnings:
        header += format_as_ul(''.join(warnings))

    return header


def format_warning(w, value):
    if w == 'no submission':
        return '<li>No submission found.</li>'

    elif w == 'unmerged branches' and value:
        branches = ['<li>{}</li>'.format(b) for b in value]
        return '<li>Repository has unmerged branches:<ul>{}</ul></li>'.format('\n'.join(branches))

    elif value:
        return '<li>Warning: {}</li>'.format(value)

    else:
        return ''


def format_file(filename, file_info):
    contents = format_file_contents(file_info.get('contents', ''), file_info) + '\n'
    compilation = format_file_compilation(file_info.get('compilation', [])) + '\n'
    test_results = format_file_results(file_info.get('result', [])) + '\n'

    if file_info.get('last modified', None):
        last_modified = ' (last modified on {})'.format(file_info['last modified'])
    else:
        last_modified = ''

    file_header = '<h2><code>{}</code>{}</h2>'.format(filename, last_modified)

    if file_info['missing']:
        note = '<code>{}</code> was not found. We found these files:\n'.format(filename)
        directory_listing = format_as_ul(['<li><code>{}</code></li>'.format(f) for f in file_info.get('other files', [])])

        if file_info['optional']:
            file_header = file_header.strip()
            file_header += ' (<strong>optional submission</strong>)\n'

        return '\n'.join([file_header, note, directory_listing + '\n\n'])

    return '\n'.join([file_header, contents, compilation, test_results])


def format_file_contents(contents, info):
    return format_as_code(html.escape(contents))


def format_file_compilation(compilations):
    result = []
    for status in compilations:
        output = status['output']
        command = '<code>{command}</code>'.format_map(status)

        if not output:
            result.append('<p><strong>no warnings: {}</strong></p>'.format(command))
        else:
            result.append('<p><strong>warnings: {}</strong></p>'.format(command))
            result.append(format_as_code(output))

    return '\n'.join(result)


def format_file_results(test_results):
    result = ''

    for test in test_results:
        header = '<p><strong>results of <code>{command}</code></strong> (status: {status})</p>'.format_map(test)
        output = format_as_code(test['output'])
        result += header + '\n' + output

        if test['truncated']:
            result += '<p><em>(truncated after {truncated after})</em></p>'.format_map(test)

    return result


def format_as_code(data):
    if not data:
        return ''
    return '<pre><code>{}</code></pre>'.format(data)


def format_as_ul(data):
    if not data:
        return ''
    return '<ul>{}</ul>'.format(''.join(data))



"""Make a nice table from the student results"""
import re
from typing import Dict, List, Tuple, TYPE_CHECKING
from termcolor import colored

from ..process_assignment.assignment_status import AssignmentStatus
from ..process_assignment.assignment_type import get_assignment_number

if TYPE_CHECKING:
    from ..student.student_result import StudentResult

COL = '|'
ROW = '-'
JOIN = '+'
MISSING = '-'
ANSI_ESCAPE = re.compile(r'\x1b[^m]*m')


def sort_by_hw_count(user: 'StudentResult') -> int:
    """Sort students by the number of completed homeworks"""
    return sum([1 if stat is AssignmentStatus.SUCCESS else 0 for _, stat in user.assignments().items()])


def sort_by_username(user: 'StudentResult') -> str:
    """Sort students by their username"""
    return user.name


def asciiify(table: str) -> str:
    """Take a flashy unicode table and render it with ASCII-only chars"""
    return ANSI_ESCAPE.sub('', table)


def pad(string: str, index: int) -> str:
    """Pad a string to the width of the stringified number"""
    padding_char = string if string == MISSING else ' '
    return string.ljust(len(str(index)), padding_char)


def symbol(assignment: Tuple[int, AssignmentStatus],
           highlight_partials: bool = True) -> str:
    """Turn an assignment status into the symbol for the table"""
    (num, status) = assignment

    if status is AssignmentStatus.SUCCESS:
        return str(num)
    elif status is AssignmentStatus.PARTIAL:
        return_val = str(num)
        if highlight_partials:
            return colored(return_val, 'red', attrs={'bold': True})
        return return_val
    else:
        return MISSING


def concat(assignments: Dict[str, AssignmentStatus],
           to_num: int,
           highlight_partials: bool = True) -> str:
    """Create the informative row of data for a list of assignment statuses"""
    nums = {get_assignment_number(a_id): (get_assignment_number(a_id), stat)
            for a_id, stat in assignments.items()}
    lst = [pad(symbol(nums[idx], highlight_partials), idx)
           if idx in nums
           else pad('-', idx)
           for idx in range(1, to_num + 1)]
    return ' '.join(lst)


def find_columns(num: int) -> str:
    """Build the table headings for the assignment sections"""
    return ' '.join([str(i) for i in range(1, num + 1)])


def columnize(student: 'StudentResult',
              longest_user: str,
              max_hwk_num: int,
              max_lab_num: int,
              max_wst_num: int,
              highlight_partials: bool = True):
    """Build the data for each row of the information table"""
    name = '{0:<{1}}'.format(student.name, len(longest_user))

    if len(student.unmerged_branches) > 0:
        name = colored(name, attrs={'bold': True})

    homework_row = concat(student.homeworks, max_hwk_num, highlight_partials)
    lab_row = concat(student.labs, max_lab_num, highlight_partials)
    worksheet_row = concat(student.worksheets, max_wst_num, highlight_partials)

    if student.error:
        return '{name}  {sep} {err}'.format(
            name=name,
            sep=COL,
            err=student.error)
    else:
        return '{name}  {sep} {hws} {sep} {labs} {sep} {wkshts}'.format(
            name=name,
            hws=homework_row,
            labs=lab_row,
            wkshts=worksheet_row,
            sep=COL)


def get_nums(students: List['StudentResult']) -> Tuple[int, int, int]:
    """Given a list of students, return the highest hw and lab number among them"""
    homework_nums = [get_assignment_number(hw) for s in students for hw in s.homeworks.keys()]
    lab_nums = [get_assignment_number(lab) for s in students for lab in s.labs.keys()]
    worksheet_nums = [get_assignment_number(ws) for s in students for ws in s.worksheets.keys()]

    max_hwk_num = max(homework_nums, default=0)
    max_lab_num = max(lab_nums, default=0)
    max_worksheet_num = max(worksheet_nums, default=0)

    return max_hwk_num, max_lab_num, max_worksheet_num


def tabulate(student_results: List['StudentResult'],
             sort_by: str = 'name',
             highlight_partials: bool = True) -> str:
    """Actually build the table"""

    # be sure that the longest username will be at least 4 chars
    usernames = [user.name for user in student_results] + ['USER']
    longest_user = max(usernames, key=len)

    # build the header row of the table
    max_hwk_num, max_lab_num, max_wst_num = get_nums(student_results)
    header_hw_nums = find_columns(max_hwk_num)
    header_lab_nums = find_columns(max_lab_num)
    header_wst_nums = find_columns(max_wst_num)
    header = '{name:<{namesize}}  {sep} {hwnums} {sep} {labnums} {sep} {wstnums}'.format(
        name='USER',
        namesize=len(longest_user),
        hwnums=header_hw_nums,
        labnums=header_lab_nums,
        wstnums=header_wst_nums,
        sep=COL)

    # build the header's bottom border
    border = ''.join([
        ''.ljust(len(longest_user) + 2, ROW),
        JOIN,
        ''.ljust(len(header_hw_nums) + 2, ROW),
        JOIN,
        ''.ljust(len(header_lab_nums) + 2, ROW),
        JOIN,
        ''.ljust(len(header_wst_nums) + 1, ROW),
    ])

    # build the table body
    # Sorts by sorter2, then by sorter1
    # This works because sorted is "stable", meaning it preserves the original order
    # So sorter1 is the main sort, and any duplicate keys are sorted by sorter2
    if sort_by == 'count':
        sorter1 = sort_by_hw_count
        sorter2 = sort_by_username
        should_reverse = True
    else:
        sorter1 = sort_by_username
        sorter2 = sort_by_hw_count
        should_reverse = False

    lines = [columnize(student, longest_user, max_hwk_num, max_lab_num, max_wst_num,
                       highlight_partials=highlight_partials)
             for student in sorted(sorted(student_results, key=sorter2), reverse=should_reverse, key=sorter1)]

    # and make the table to return
    table = [header, border] + lines
    return '\n'.join(table)



from pyfakefs import fake_filesystem
from stograde.common.chdir import chdir


def test_chdir(fs):
    os_module = fake_filesystem.FakeOsModule(fs)

    fs.create_dir('./folder')

    assert os_module.getcwd() == '/'
    with chdir('./folder'):
        assert os_module.getcwd() == '/folder'
    assert os_module.getcwd() == '/'



from bidict import bidict
import sys

from ..common import chdir, run
from ..common.run_status import RunStatus
from ..toolkit import global_vars

SPEC_URLS = bidict({
    'sd': 'https://github.com/StoDevX/cs251-specs.git',
    'hd': 'https://github.com/StoDevX/cs241-specs.git',
    'ads': 'https://github.com/StoDevX/cs253-specs.git',
    'os': 'https://github.com/StoDevX/cs273-specs.git'
})


def download_specs(course: str, basedir: str):
    course = course.split("/")[0].lower()
    try:
        url = SPEC_URLS[course]
    except KeyError:
        print("Course {} not recognized".format(course))
        sys.exit(1)
    with chdir(basedir):
        print('Downloading specs for {}'.format(course.upper()))
        status, _, _ = run(['git', 'clone', url, 'data'])
        if status is RunStatus.SUCCESS:
            print('Download complete')
        else:
            print('Download failed: {}'.format(status.name))
            sys.exit(1)


def create_data_dir(course: str, basedir: str):
    if global_vars.CI:
        if course:
            download_specs(course, basedir)
        else:
            print("data directory not found and no course specified")
            sys.exit(1)

    else:
        print('data directory not found', file=sys.stderr)
        if course:
            download_specs(course, basedir)
        else:
            download = input("Download specs? (y/N) ")
            if download and download.lower()[0] == "y":
                repo = input("Which class? (SD/HD/ADS/OS) ")
                if repo:
                    download_specs(repo, basedir)
                else:
                    sys.exit(1)
            else:
                print('Not downloading specs')
                sys.exit(1)


def get_supported_courses() -> str:
    course_list = [course for course in SPEC_URLS.keys()]
    courses = ''
    for course in course_list:
        courses += course + ', '
    courses = courses[:-2]
    return courses



from .analyze_student import analyze_student
from .checkout import checkout_date, checkout_ref
from .ci_analyze import ci_analyze
from .clone import clone_student, clone_url
from .process_student import process_student, prepare_student_repo
from .pull import pull
from .record_student import record_student
from .remove import remove
from .reset import reset
from .stash import stash



import smtplib
import os


def send_email(msg):
    username = os.getenv('STOGRADE_EMAIL_USERNAME', None)
    if not username:
        raise Exception('Missing $STOGRADE_EMAIL_USERNAME')

    password = os.getenv('STOGRADE_EMAIL_PASSWORD', None)
    if not password:
        raise Exception('Missing $STOGRADE_EMAIL_PASSWORD')

    # Send the message via our own SMTP server.
    with smtplib.SMTP_SSL('smtp.gmail.com', 465) as s:
        s.login(username, password)
        s.send_message(msg)



import pytest
from stograde.formatters.truncate import truncate


def test_truncate():
    assert truncate("a long string", 5) == "a lon"
    assert truncate("a", 5) == "a"

    assert truncate("😀😀😀", 4) == "😀"
    assert truncate("😀😀😀", 8) == "😀😀"
    assert truncate("😀😀😀", 12) == "😀😀😀"

    assert truncate("абвгд", 5) == "аб"

    try:
        truncate("😀😀😀", 1)
    except UnicodeDecodeError:
        pytest.fail("Unexpected UnicodeDecodeError")

    assert truncate("👨‍👨‍👧‍👧", 4) == "👨"
    assert truncate("👨‍👨‍👧‍👧", 7) == "👨‍"



"""Post a private gist of the analysis to github's gist service"""
import getpass
import json
import requests
__all__ = ['post_gist']


def get_auth():
    """Get the user's credentials"""
    username = input('Github username: ')
    password = getpass.getpass('Github password: ')
    return username, password


def post_gist(description, files):
    """Post a gist of the analysis"""
    username, password = get_auth()
    sess = requests.Session()
    sess.auth = (username, password)

    params = {
        'description': description,
        'files': files,
        'public': False,
    }

    headers = {
        'Content-Type': 'application/json',
        'Accept': '*/*',
        'User-Agent': 'stolaf-cs-toolkit/v1',
    }

    req = sess.post('https://api.github.com/gists',
                    headers=headers,
                    data=json.dumps(params))

    result = req.json()

    return result.get('html_url', '"' + result.get('message', 'Error') + '"')



from stograde.common.run import run
from stograde.common.run_status import RunStatus


def test_run():
    status, result, again = run(['echo', 'hi'])
    assert status == RunStatus.SUCCESS
    assert result == 'hi\n'
    assert again is False


def test_run_stdin():
    status, result, again = run(['cat'], input_data=b'hello')
    assert status == RunStatus.SUCCESS
    assert result == 'hello'
    assert again is False


def test_run_timeout():
    status, result, again = run(['sleep', '1'], timeout=0.5)
    assert status == RunStatus.TIMEOUT_EXPIRED
    assert result == "Command '['sleep', '1']' timed out after 0.5 seconds"
    assert again is False


def test_run_not_found():
    status, result, again = run(['notfound'])
    assert status == RunStatus.FILE_NOT_FOUND
    # TODO: fix this on python 3.6
    # assert result == "[Errno 2] No such file or directory: 'notfound'"
    assert again is False



import os
from typing import TYPE_CHECKING

from .compile_result import CompileResult
from .file_result import FileResult
from .test_result import TestResult
from ..common import cat, run, pipe
from ..common.run_status import RunStatus
from ..formatters.truncate import truncate

if TYPE_CHECKING:
    from ..specs.spec import SpecFile


def get_file(file_spec: 'SpecFile', file_result: FileResult) -> bool:
    file_status, file_contents = cat(file_spec.file_name)
    if file_status is RunStatus.SUCCESS:
        _, last_edit, _ = run(['git', 'log', '-n', '1', '--pretty=format:%cd', '--', file_spec.file_name])
        file_result.last_modified = last_edit

    if file_spec.options.hide_contents:
        file_contents = ''
    elif file_spec.options.truncate_contents:
        file_contents = truncate(file_contents, file_spec.options.truncate_contents)

    if file_status is not RunStatus.SUCCESS:
        file_result.file_missing = True
        file_result.other_files = os.listdir('.')
        file_result.optional = file_spec.options.optional
        return False
    else:
        file_result.compile_optional = file_spec.options.compile_optional
        file_result.contents = file_contents
        return True


def compile_file(*, file_spec: 'SpecFile', results: FileResult, supporting_dir: str) -> bool:
    for command in file_spec.compile_commands:
        command = command \
            .replace('$@', './' + file_spec.file_name) \
            .replace('$SUPPORT', supporting_dir)

        cmd, input_for_cmd = pipe(command)
        status, compile_output, _ = run(cmd, timeout=30, input_data=input_for_cmd)

        results.compile_results.append(CompileResult(command=command,
                                                     output=compile_output,
                                                     status=status))

        if status != 'success':
            return False

    return True


def test_file(*,
              file_spec: 'SpecFile',
              file_results: FileResult,
              cwd: str,
              supporting_dir: str,
              interact: bool):
    for test_cmd in file_spec.test_commands:
        if not test_cmd:
            continue

        test_cmd = test_cmd \
            .replace('$@', './' + file_spec.file_name) \
            .replace('$SUPPORT', supporting_dir)

        test_cmd, input_for_test = pipe(test_cmd)

        if os.path.exists(os.path.join(cwd, file_spec.file_name)):
            again = True
            while again:
                status, full_result, again = run(test_cmd,
                                                 input_data=input_for_test,
                                                 timeout=file_spec.options.timeout,
                                                 interact=interact)

                result = truncate(full_result, file_spec.options.truncate_output)
                was_truncated = (full_result != result)

                file_results.test_results.append(TestResult(
                    command=test_cmd,
                    output=result,
                    status=status,
                    error=status != RunStatus.SUCCESS,
                    truncated=was_truncated,
                    truncated_after=file_spec.options.truncate_output,
                ))

        else:
            file_results.test_results.append(TestResult(
                command=test_cmd,
                output='{} could not be found.'.format(file_spec.file_name),
                status=RunStatus.FILE_NOT_FOUND,
                error=True,
            ))


def process_file(*,
                 file_spec: 'SpecFile',
                 cwd: str,
                 supporting_dir: str,
                 interact: bool,
                 skip_web_compile: bool) -> FileResult:
    file_result = FileResult(file_name=file_spec.file_name)

    should_continue = get_file(file_spec, file_result)
    if not should_continue or skip_web_compile and file_spec.options.web_file:
        return file_result

    should_continue = compile_file(file_spec=file_spec,
                                   results=file_result,
                                   supporting_dir=supporting_dir)

    if not should_continue or file_spec.options.web_file:
        return file_result

    test_file(file_spec=file_spec,
              file_results=file_result,
              cwd=cwd,
              supporting_dir=supporting_dir,
              interact=interact)

    return file_result



import datetime
import os
import re
import sys

from ..common import chdir, run
from ..specs import SPEC_URLS

COURSE_REGEX = re.compile(r'^([\w]{2,3}/[sf]\d\d)$')


def compute_stogit_url(*, stogit: str, course: str, _now: datetime.date) -> str:
    """calculate a default stogit URL, or use the specified one"""
    if stogit:
        return stogit
    elif re.match(COURSE_REGEX, course):
        return 'git@stogit.cs.stolaf.edu:{}'.format(course)
    else:
        if not course:
            course = get_course_from_specs()
            print('Course {} inferred from specs'.format(course.upper()), file=sys.stderr)
        semester = 's' if _now.month < 7 else 'f'
        year = str(_now.year)[2:]
        return 'git@stogit.cs.stolaf.edu:{}/{}{}'.format(course, semester, year)


def get_course_from_specs() -> str:
    if not os.path.exists("data"):
        print('Cannot determine course from specs', file=sys.stderr)
        sys.exit(1)

    with chdir('./data'):
        _, res, _ = run(['git', 'config', '--get', 'remote.origin.url'])
        try:
            course = SPEC_URLS.inverse[res.rstrip()]
        except KeyError:
            course = 'sd'  # default to SD as last resort
        finally:
            return course



from stograde.student import remove
from stograde.student import clone_url
from stograde.student import checkout_ref
from stograde.student import record_student
from stograde.student import analyze_student


def process_student(*, repo, branch, assignments, folder, specs, basedir, debug=False):
    clone_url(repo, into=folder)

    try:
        # this is usually going to be a no-op (for any commits on master)
        checkout_ref(folder, ref=branch)

        recordings = record_student(folder, specs=specs, assignments=assignments, basedir=basedir, interact=False)
        analysis = analyze_student(folder, specs, check_for_branches=False)

        remove(folder)

        return analysis, recordings

    except Exception as err:
        if debug:
            raise err
        return {'username': folder, 'error': err}, []



from inspect import isgenerator
from stograde.common.group_by import group_by


def test_group_by():
    assert isgenerator(group_by([1, 2, 3], lambda s: s % 2 == 0))
    assert dict(group_by(['1', '2', '3'], lambda s: s.isdigit())) == {True: ['1', '2', '3']}
    assert dict(group_by(
        [{'a': 1}, {'a': 2}, {'a': 1}],
        lambda i: i['a'])) == {1: [{'a': 1}, {'a': 1}], 2: [{'a': 2}]}



import textwrap
import sys
import os

import pytest

from stograde.toolkit.__main__ import main

_dir = os.path.dirname(os.path.realpath(__file__))


@pytest.mark.datafiles(os.path.join(_dir, 'fixtures', 'two_students_hw1'))
def test_stograde_table(datafiles, capsys):
    os.chdir(str(datafiles))

    argv = sys.argv
    sys.argv = [argv[0]] + ['table', '--skip-repo-update', '--skip-spec-update', '--skip-version-check',
                            '--skip-dependency-check', '--no-partials']

    try:
        main()
    except SystemExit:
        pass

    out, err = capsys.readouterr()

    assert out == textwrap.dedent("\n"
                                  "USER      | 1 | 1 | 1\n"
                                  "----------+---+---+--\n"
                                  "rives     | - | - | -\n"
                                  "student2  | 1 | 1 | -\n\n")

    sys.argv = argv


@pytest.mark.datafiles(os.path.join(_dir, 'fixtures', 'two_students_hw1'))
def test_stograde_hidden_table(datafiles, capsys):
    os.chdir(str(datafiles))

    argv = sys.argv
    sys.argv = [argv[0]] + ['record', 'hw1', '--skip-repo-update', '--skip-spec-update', '--skip-version-check',
                            '--skip-dependency-check']

    try:
        main()
    except SystemExit:
        pass

    out, err = capsys.readouterr()

    assert out == ""

    sys.argv = argv


@pytest.mark.datafiles(os.path.join(_dir, 'fixtures', 'two_students_hw1'))
def test_stograde_record(datafiles):
    os.chdir(str(datafiles))

    argv = sys.argv
    sys.argv = [argv[0]] + ['record', 'hw1', '--skip-repo-update', '--skip-spec-update', '--skip-version-check',
                            '--skip-dependency-check']

    try:
        main()
    except SystemExit:
        pass

    assert (datafiles / 'logs' / 'log-hw1.md').isfile()

    sys.argv = argv



from .download_specs import create_data_dir, get_supported_courses, SPEC_URLS
from .filter_specs import filter_assignments, find_all_specs
from .load import load_specs
from .util import check_architecture, check_spec_dependencies, get_filenames



from dataclasses import dataclass, field
from typing import List, TYPE_CHECKING

from .submission_warnings import SubmissionWarnings

if TYPE_CHECKING:
    from ..process_file.file_result import FileResult


@dataclass
class RecordResult:
    """The results of testing a student's submission for an assignment"""
    spec_id: str  # The spec being tested against
    student: str  # The student's username
    first_submission: str = ''  # The first commit of this assignment's files
    warnings: SubmissionWarnings = field(default_factory=SubmissionWarnings)  # Any warnings about the assignment
    file_results: List['FileResult'] = field(default_factory=list)  # The results from each individual file



import logging
import sys
from os import path

from ..common import run
from ..common.run_status import RunStatus


def clone_student(student: str, base_url: str):
    logging.debug("Cloning {}'s repository".format(student))
    if not path.exists(student):
        clone_url('{}/{}.git'.format(base_url, student))


def clone_url(url: str, into: str = None):
    if into:
        logging.info('cloning {} into {}'.format(url, into))
        status, output, _ = run(['git', 'clone', '--quiet', url, into])
    else:
        logging.info('cloning {}'.format(url))
        status, output, _ = run(['git', 'clone', '--quiet', url])
    logging.debug(output)
    if status is RunStatus.CALLED_PROCESS_ERROR and 'Permission denied' in output:
        print('Permission denied when cloning from {}.', file=sys.stderr)
        print('Make sure that this SSH key is registered with StoGit.'.format(url), file=sys.stderr)
        sys.exit(1)



import datetime
from stograde.toolkit.args import build_argparser

from stograde.toolkit.stogit_url import compute_stogit_url


def args(arglist):
    return vars(build_argparser().parse_args(args=arglist))


def test_stogit_url_computation():
    assert compute_stogit_url(stogit='', course='sd', _now=datetime.date(2017, 1, 31)) \
        == 'git@stogit.cs.stolaf.edu:sd/s17'

    assert compute_stogit_url(stogit='', course='sd', _now=datetime.date(2016, 9, 15)) \
        == 'git@stogit.cs.stolaf.edu:sd/f16'

    assert compute_stogit_url(stogit='', course='sd', _now=datetime.date(2016, 4, 15)) \
        == 'git@stogit.cs.stolaf.edu:sd/s16'

    assert compute_stogit_url(stogit='blah', course='sd', _now=datetime.date.today()) \
        == 'blah'

    assert compute_stogit_url(stogit='', course='hd', _now=datetime.date(2016, 4, 15)) \
        == 'git@stogit.cs.stolaf.edu:hd/s16'



import functools
import logging
import os
from os import makedirs
import sys
from threading import Thread
from typing import Any, Dict, List, TYPE_CHECKING

from . import global_vars
from .process_repos import process_parallel_repos
from .process_students import process_students
from .save_recordings import save_recordings
from .tabulate import tabulate
from ..common import chdir
from ..student import ci_analyze, prepare_student_repo
from ..webapp import is_web_spec, launch_cli, server

if TYPE_CHECKING:
    from ..specs.spec import Spec
    from ..student.student_result import StudentResult


def do_ci(specs: Dict[str, 'Spec'],
          students: List[str],
          base_dir: str,
          stogit_url: str,
          args: Dict[str, Any]):
    global_vars.CI = True
    skip_web_compile: bool = args['skip_web_compile']

    results: List['StudentResult'] = process_students(specs=specs,
                                                      students=students,
                                                      analyze=True,
                                                      base_dir=base_dir,
                                                      clean=False,
                                                      date='',
                                                      interact=False,
                                                      no_progress_bar=True,
                                                      record=True,
                                                      skip_branch_check=True,
                                                      skip_repo_update=True,
                                                      skip_web_compile=skip_web_compile,
                                                      stogit_url=stogit_url,
                                                      workers=1,
                                                      work_dir='./students')

    passing: bool = ci_analyze(results)
    if not passing:
        logging.debug('Build failed')
        sys.exit(1)


def do_record(specs: Dict[str, 'Spec'],
              students: List[str],
              base_dir: str,
              stogit_url: str,
              args: Dict[str, Any]):
    show_table: bool = args['table']

    clean: bool = args['clean']
    date: str = args['date']
    gist: bool = args['gist']
    interact: bool = args['interact']
    no_partials: bool = args['no_partials']
    no_progress_bar: bool = args['no_progress_bar']
    skip_branch_check: bool = args['skip_branch_check']
    skip_repo_update: bool = args['skip_repo_update']
    skip_web_compile: bool = args['skip_web_compile']
    sort_by: str = args['sort_by']
    workers: int = args['workers'] if not global_vars.DEBUG and not interact else 1

    makedirs('./students', exist_ok=True)

    results: List['StudentResult'] = process_students(specs=specs,
                                                      students=students,
                                                      analyze=show_table,
                                                      base_dir=base_dir,
                                                      clean=clean,
                                                      date=date,
                                                      interact=interact,
                                                      no_progress_bar=no_progress_bar,
                                                      record=True,
                                                      skip_branch_check=skip_branch_check,
                                                      skip_repo_update=skip_repo_update,
                                                      skip_web_compile=skip_web_compile,
                                                      stogit_url=stogit_url,
                                                      workers=workers,
                                                      work_dir='./students')

    table: str = ''
    if show_table:
        table = tabulate(results, sort_by=sort_by, highlight_partials=not no_partials)
        print('\n' + table + '\n')

    save_recordings(results, table, gist=gist)


def do_table(specs: Dict[str, 'Spec'],
             students: List[str],
             base_dir: str,
             stogit_url: str,
             args: Dict[str, Any]):
    clean: bool = args['clean']
    date: str = args['date']
    no_partials: bool = args['no_partials']
    no_progress_bar: bool = args['no_progress_bar']
    skip_repo_update: bool = args['skip_repo_update']
    sort_by: str = args['sort_by']
    workers: int = args['workers'] if not global_vars.DEBUG else 1

    results: List['StudentResult'] = process_students(specs=specs,
                                                      students=students,
                                                      analyze=True,
                                                      base_dir=base_dir,
                                                      clean=clean,
                                                      date=date,
                                                      interact=False,
                                                      no_progress_bar=no_progress_bar,
                                                      record=False,
                                                      skip_branch_check=True,
                                                      skip_repo_update=skip_repo_update,
                                                      skip_web_compile=True,
                                                      stogit_url=stogit_url,
                                                      workers=workers,
                                                      work_dir='./students')

    print('\n' + tabulate(results, sort_by=sort_by, highlight_partials=not no_partials) + '\n')


def do_web(specs: Dict[str, 'Spec'],
           students: List[str],
           base_dir: str,
           stogit_url: str,
           args: Dict[str, Any]):
    clean: bool = args['clean']
    date: str = args['date']
    no_progress_bar: bool = args['no_progress_bar']
    skip_repo_update: bool = args['skip_repo_update']
    workers: int = args['workers'] if not global_vars.DEBUG else 1
    port: int = args['port']
    spec: 'Spec' = list(specs.values())[0]

    if not is_web_spec(spec):
        print("No web files in assignment {}".format(spec.id))
        sys.exit(1)

    Thread(target=server.run_server, args=(port,), daemon=True).start()

    launch_cli(base_dir=base_dir,
               clean=clean,
               date=date,
               no_progress_bar=no_progress_bar,
               skip_repo_update=skip_repo_update,
               spec=spec,
               stogit_url=stogit_url,
               students=students,
               workers=workers)


def do_clean(students: List[str],
             stogit_url: str,
             base_dir: str,
             no_progress_bar: bool,
             workers: int):
    with chdir(os.path.join(base_dir, 'students')):
        single_repo = functools.partial(
            prepare_student_repo,
            stogit_url=stogit_url,
            do_clean=True,
            do_clone=True,
            do_pull=True,
            do_checkout=False
        )

        process_parallel_repos(students=students,
                               no_progress_bar=no_progress_bar,
                               workers=workers,
                               operation=single_repo)


def do_update(students: List[str],
              stogit_url: str,
              base_dir: str,
              no_progress_bar: bool,
              workers: int):
    with chdir(os.path.join(base_dir, 'students')):
        single_repo = functools.partial(
            prepare_student_repo,
            stogit_url=stogit_url,
            do_clean=False,
            do_clone=True,
            do_pull=True,
            do_checkout=False
        )

        process_parallel_repos(students=students,
                               no_progress_bar=no_progress_bar,
                               workers=workers,
                               operation=single_repo)



from .base import format_collected_data
from .html import format_assignment_html as html
from .markdown import format_assignment_markdown as markdown
from .truncate import truncate
from .yaml import format_assignment_yaml as yaml



import natsort
import re
import requests

from .config import conf
from ..common import version


def get_all_versions(pkg='stograde'):
    # PyPI has these "simple" html pages. They're how pip does stuff.
    try:
        req = requests.get('https://pypi.python.org/simple/{}'.format(pkg), timeout=0.5)
    except requests.exceptions.ConnectionError:
        return []
    except requests.exceptions.Timeout:
        return []

    # Remove the first and last bits
    page = re.sub(r'.*</h1>|<br/>.*', '', req.text)
    # Grab just the links from the page
    lines = [line for line in page.splitlines() if line.strip().startswith('<a')]
    # Grab just the middles of the links
    packages = [re.sub('.*>(.*)<.*', '\\1', line) for line in lines]
    # Remove the suffixes
    versions = [re.sub(r'(\d)-.*|\.tar.gz', '\\1', line) for line in packages]
    # Remove the prefixes
    versions = [re.sub(r'.*-(\d)', '\\1', line) for line in versions]
    # Return the sorted list of all available versions
    return natsort.natsorted(set(versions))


def update_available():
    if not conf.needs_update_check():
        return version, None

    conf.set_last_update_check()

    all_versions = get_all_versions()

    if version not in all_versions:
        return version, None
    if all_versions.index(version) != len(all_versions) - 1:
        return version, all_versions[-1]

    return version, None



import re


def clean_header(text):
    return text.strip('[]')


def get_students():
    try:
        with open('students.txt', 'r', encoding='utf-8') as infile:
            lines = infile.readlines()
    except FileNotFoundError:
        lines = []

    lines = [l.strip() for l in lines]
    lines = [l for l in lines if l]
    group_regex = re.compile(r'\[.*\]')
    group_list = [clean_header(l) for l in lines if group_regex.match(l)]
    if not group_list:
        group_list = ['my']

    groups = {group: [] for group in group_list}
    current_group = group_list[0]
    for line in lines:
        if group_regex.match(line):
            current_group = clean_header(line)
        else:
            groups[current_group].append(line)

    return groups



from typing import TYPE_CHECKING

from ..common import find_unmerged_branches_in_cwd

if TYPE_CHECKING:
    from ..student.student_result import StudentResult


def find_unmerged_branches(result: 'StudentResult'):
    """Find any unmerged branches and add them to the result"""
    # approach taken from https://stackoverflow.com/a/3602022/2347774
    unmerged_branches = find_unmerged_branches_in_cwd()
    if unmerged_branches:
        result.unmerged_branches = unmerged_branches



from collections import defaultdict
from typing import List, TYPE_CHECKING

from ..common.group_by import group_by as group

if TYPE_CHECKING:
    from ..student.student_result import StudentResult


def format_collected_data(student_results: List['StudentResult'],
                          group_by: str,
                          formatter) -> defaultdict:
    """Turn the list of recordings into a list of nicely-formatted results.

    `grouped_records` will be a list of pairs: (assignment, recordings), where
    `assignment` is the assignment name and `recordings` is a list of recordings
    (one per student).

    `formatter` is a formatter. It receives each recording, one at a time, and should
    return a {content: str, student: str, type: str, assignment: str} dict for each
    recording.
    """

    results = []
    for student in student_results:
        results.extend(student.results)

    if group_by == 'assignment':
        grouped_records = group(results, lambda rec: rec.spec_id)
    elif group_by == 'student':
        grouped_records = group(results, lambda rec: rec.student)
    else:
        # not entirely sure what this'll do
        grouped_records = results

    student_results = defaultdict(list)
    for key, recordings in grouped_records:
        for content in recordings:
            student_results[key].append(formatter(content))

    return student_results



import logging
import sys

from ..common import chdir, run
from ..common.run_status import RunStatus


def pull(student: str):
    logging.debug("Pulling {}'s repository".format(student))
    with chdir(student):
        status, output, _ = run(['git', 'pull', '--quiet', 'origin', 'master'])

    if status is RunStatus.CALLED_PROCESS_ERROR and 'not a git repository' in output:
        print('Student directory {} not a git repository', file=sys.stderr)
        print('Try running "stograde repo reclone"'.format(student), file=sys.stderr)



import logging
from typing import Optional

from ..common import chdir, run


def checkout_date(student: str, date: Optional[str] = None):
    if date:
        logging.debug("Checking out commits in {}'s repository before {}".format(student, date))
        with chdir(student):
            _, rev, _ = run(['git', 'rev-list', '-n', '1', '--before="{} 18:00"'.format(date), 'master'])
            run(['git', 'checkout', rev, '--force', '--quiet'])


def checkout_ref(student: str, ref: str):
    with chdir(student):
        run(['git', 'checkout', ref, '--force', '--quiet'])



from enum import auto, Enum
import re
from typing import Tuple


class AssignmentType(Enum):
    """The type of assignment"""
    HOMEWORK = auto()
    LAB = auto()
    WORKSHEET = auto()


def get_assignment_type(name: str) -> AssignmentType:
    """Determine the assignment type from a string

    Throws a ValueError if name is not a valid assignment type"""
    kind, _ = parse_assignment_name(name)

    if kind == 'hw':
        return AssignmentType.HOMEWORK
    elif kind == 'lab':
        return AssignmentType.LAB
    elif kind == 'ws':
        return AssignmentType.WORKSHEET
    else:
        raise ValueError("Could not parse assignment type for {}".format(name))


def get_assignment_number(name: str) -> int:
    """Returns the number associated with the assignment"""
    _, num = parse_assignment_name(name)
    return int(num)


def parse_assignment_name(name: str) -> Tuple[str, str]:
    """Splits an assignment id into its type and number"""
    matches = re.match(r'([a-zA-Z]+)(\d+)', name).groups()
    return matches[0], matches[1]



from textwrap import indent
import traceback
from typing import Dict, List

from ..process_assignment.record_result import RecordResult
from ..process_assignment.submission_warnings import SubmissionWarnings
from ..process_file.compile_result import CompileResult
from ..process_file.file_result import FileResult
from ..process_file.test_result import TestResult
from ..toolkit import global_vars


def format_assignment_markdown(result: RecordResult) -> Dict:
    """Given a single recording, format it into a markdown file.

    Each recording will only have one student.

    Returns a {content: str, student: str, type: str, assignment: str} dict.
    """

    try:
        files = format_files_list(result.file_results)
        warnings = format_warnings(result.warnings)
        header = format_header(result, warnings)
        output = (header + files) + '\n\n'

    except Exception as err:
        if global_vars.DEBUG:
            raise err
        output = indent(traceback.format_exc(), ' ' * 4) + '\n\n'

    return {
        'assignment': result.spec_id,
        'content': output,
        'student': result.student,
        'type': 'md',
    }


def format_files_list(files) -> str:
    return '\n\n' + '\n\n'.join([format_file(info) for info in files])


def format_header(result: RecordResult, warnings: str) -> str:
    """Format the header for the section of the log file"""

    header = '# {spec} – {student}\n{first_submit}\n'.format(spec=result.spec_id,
                                                             student=result.student,
                                                             first_submit=result.first_submission)

    if warnings:
        header += '\n' + warnings + '\n'

    return header


def format_warnings(warnings: SubmissionWarnings) -> str:
    if warnings.assignment_missing:
        return '**No submission found**\n'

    elif warnings.unmerged_branches:
        branches = ['  - ' + b for b in warnings.unmerged_branches]
        return '### *Repository has unmerged branches:*\n{}'.format('\n'.join(branches))

    elif warnings.recording_err:
        return '**Warning: ' + warnings.recording_err + '**'

    else:
        return ''


def format_file(file_info: FileResult) -> str:
    """Format a file for the log.
    Formats and concatenates a header, the file contents, compile output and test output.

    Last modification is calculated and added to header.
    If file does not exist, adds a list of all files in the directory.
    If file is missing and is optional, adds a note in place of last modification time.
    """

    contents = format_file_contents(file_info.contents, file_info.file_name) + '\n'
    compilation = format_file_compilation(file_info.compile_results) + '\n'
    test_results = format_file_results(file_info.test_results) + '\n'

    if file_info.last_modified:
        last_modified = ' ({})'.format(file_info.last_modified)
    else:
        last_modified = ''

    file_header = '## {}{}\n'.format(file_info.file_name, last_modified)

    if file_info.file_missing:
        note = 'File not found. `ls .` says that these files exist:\n'
        directory_listing = indent('\n'.join(file_info.other_files), ' ' * 4)

        if file_info.optional:
            file_header = file_header.strip()
            file_header += ' (**optional submission**)\n'

        return '\n'.join([file_header, note, directory_listing + '\n\n'])

    return '\n'.join([file_header, contents, compilation, test_results])


def format_file_contents(contents: str, filename: str) -> str:
    """Add markdown code block around file contents with extension for code highlighting.

    If a file is empty or contains only whitespace, note this in the log.
    """

    if not contents.rstrip():
        return '*File empty*'
    return '```{}\n'.format(filename.split('.')[-1]) + contents + '\n```\n'


def format_file_compilation(compilations: List[CompileResult]) -> str:
    """Add header and markdown code block to compile command outputs"""

    result = []
    for compile_result in compilations:
        output = compile_result.output
        command = '`{command}`'.format(command=compile_result.command)

        if not output:
            result.append('**no warnings: {}**\n\n'.format(command))
        else:
            result.append('**warnings: {}**\n'.format(command))
            result.append('```\n' + output + '\n```\n')

    return '\n'.join(result)


def format_file_results(test_results: List[TestResult]) -> str:
    """Add header and markdown code block to test outputs"""

    result = []
    for test in test_results:
        header = '**results of `{command}`** (status: {status})\n'.format(command=test.command,
                                                                          status=test.status)
        if test.output:
            result.append(header + '\n```\n' + test.output + '\n```\n')
            if test.truncated:
                result.append('(truncated after {})'.format(test.truncated_after))
        else:
            result.append(header)

    return '\n'.join(result)



import logging

from ..common import chdir, run


def stash(student: str):
    logging.debug("Stashing {}'s repository".format(student))
    with chdir(student):
        if has_changed_files():
            run(['git', 'stash', '-u'])
            run(['git', 'stash', 'clear'])


def has_changed_files() -> bool:
    _, output, _ = run(['git', 'status', '--porcelain'])
    return bool(output)



from stograde.common.pipe import pipe, expand_chunk, process_chunk


def test_expand_chunk(fs):
    fs.create_file('file1')
    fs.create_file('file2')

    assert expand_chunk('ls') == ['ls']
    assert sorted(expand_chunk('file*')) == sorted(['file1', 'file2'])


def test_process_chunk(fs):
    fs.create_file('file1')
    fs.create_file('file2')

    assert process_chunk('echo Hawken \n 26') == ['echo', 'Hawken', '26']


def test_pipe():
    assert pipe('echo hi | cat') == (['cat'], b'hi\n')



import sys

from ..common import run
from ..common.run_status import RunStatus


def check_dependencies():
    check_stogit_known_host()
    check_git_installed()


def check_stogit_known_host():
    status, output, _ = run(['ssh-keygen', '-F', 'stogit.cs.stolaf.edu'])
    if status is RunStatus.CALLED_PROCESS_ERROR and 'exit status 1' in output:
        print('stogit.cs.stolaf.edu not in known hosts', file=sys.stderr)
        print('Run "ssh-keyscan stogit.cs.stolaf.edu >> ~/.ssh/known_hosts" to fix', file=sys.stderr)
        sys.exit(1)


def check_git_installed():
    status, output, _ = run(['git', '--version'])
    if status is RunStatus.FILE_NOT_FOUND and 'No such file or directory' in output:
        print('git is not installed', file=sys.stderr)
        print('Install git to continue', file=sys.stderr)
        sys.exit(1)



"""Deal with argument parsing for Referee"""

import argparse
import sys
import json


def get_args():
    """Construct the argument list and parse the passed arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument('data', help='data', nargs='?')
    parser.add_argument('--stdin', action='store_true', help='read from stdin')
    parser.add_argument('--debug', action='store_true', help='enable debugging mode (throw errors, etc)')
    parser.add_argument('--send', action='store_true', help='actually send emails')

    # parser.add_argument('STOGIT_URL', help='The stogit base URL')
    # parser.add_argument('USERNAME', help='Which student to process')
    # parser.add_argum?ent('COMMIT', nargs='+', help='Commit hashes to process')

    return parser


def process_args():
    """Process the arguments and create usable data from them"""
    parser = get_args()
    args = vars(parser.parse_args())
    if args['stdin']:
        args['data'] = sys.stdin.read()
    args['data'] = json.loads(args['data'])
    return args



from dataclasses import dataclass
from typing import List

from .file_options import FileOptions


@dataclass
class SpecFile:
    file_name: str
    compile_commands: List[str]
    test_commands: List[str]
    options: FileOptions

    def add_from_tests(self, test_specs: list):
        """Support legacy specs with a separate `tests:` tag

        :param test_specs: The `list` under `tests:`
        """

        test_commands = []

        for file_test_spec in test_specs:
            if isinstance(file_test_spec, dict):
                if file_test_spec.get('file') == self.file_name:
                    commands = file_test_spec.get('commands', [])
                    if isinstance(commands, str):
                        commands = [commands]
                    test_commands += commands

            elif isinstance(file_test_spec, list):  # legacy spec support
                if len(file_test_spec) == 0:
                    continue
                elif file_test_spec[0] == self.file_name:
                    test_commands += [f for f in file_test_spec[1:] if isinstance(f, str)]

            else:
                raise TypeError('Cannot parse "tests:": incorrect data type for a test: {}'
                                .format(type(file_test_spec)))

        return self


def create_spec_file(file_spec) -> SpecFile:
    """Create a new SpecFile instance

    :param file_spec: A `dict` or `list` with the file specifications
    """

    if isinstance(file_spec, dict):
        assert 'file' in file_spec
        file_name = file_spec['file']

        compile_commands = file_spec.get('commands', [])
        if isinstance(compile_commands, str):
            compile_commands = [compile_commands]
        assert isinstance(compile_commands, list)

        test_commands = file_spec.get('tests', [])
        if isinstance(test_commands, str):
            test_commands = [test_commands]
        assert isinstance(test_commands, list)

        file_options = file_spec.get('options', {})
        options = FileOptions().update(file_options)

    elif isinstance(file_spec, list):  # legacy spec support
        file_name = file_spec[0]
        compile_commands = [f for f in file_spec[1:] if isinstance(f, str)]
        test_commands = []
        option_list = [opt for opt in file_spec[1:] if isinstance(opt, dict)]
        option_dict = {k: v for opt in option_list for k, v in opt.items()}
        options = FileOptions().update(option_dict)

    else:
        raise TypeError('Cannot parse "files:": incorrect data type for a file: {}'.format(type(file_spec)))

    return SpecFile(file_name=file_name,
                    compile_commands=compile_commands,
                    test_commands=test_commands,
                    options=options)



from itertools import groupby
from types import LambdaType
from typing import Any, Generator, Iterable, List, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from ..process_assignment.record_result import RecordResult

__all__ = ['group_by']


def group_by(iterable: Iterable['RecordResult'],
             predicate: LambdaType) -> Generator[Tuple[Any, List['RecordResult']], None, None]:
    """Group an iterable by a predicate function"""
    sorted_iterable = sorted(iterable, key=predicate)
    grouped = groupby(sorted_iterable, key=predicate)
    for k, v in grouped:
        yield k, list(v)



from concurrent.futures import as_completed, ProcessPoolExecutor
import functools
import logging
import os
from typing import List

from ..toolkit.progress_bar import make_progress_bar


def process_parallel_repos(students: List[str],
                           no_progress_bar: bool,
                           workers: int,
                           operation: functools.partial):
    if workers > 1:
        print_progress = make_progress_bar(students, no_progress_bar=no_progress_bar)
        with ProcessPoolExecutor(max_workers=workers) as pool:
            futures = [pool.submit(operation, name) for name in students]
            for future in as_completed(futures):
                completed_student = future.result()
                print_progress(completed_student)
    else:
        for student in students:
            logging.debug('Processing {}'.format(student))
            operation(student)


def create_students_dir(base_dir: str):
    os.makedirs(os.path.join(base_dir, 'students'), exist_ok=True)



from typing import Dict, List, TYPE_CHECKING

from .analyze_student import analyze_student
from .record_student import record_student
from ..student import checkout_date, clone_student
from ..student.pull import pull
from ..student.remove import remove
from ..student.reset import reset
from ..student.stash import stash
from ..student.student_result import StudentResult
from ..toolkit import global_vars

if TYPE_CHECKING:
    from ..specs.spec import Spec


def process_student(
        student: str,
        *,
        analyze: bool,
        basedir: str,
        clean: bool,
        date: str,
        interact: bool,
        record: bool,
        skip_branch_check: bool,
        skip_repo_update: bool,
        skip_web_compile: bool,
        specs: Dict[str, 'Spec'],
        stogit_url: str
) -> StudentResult:
    assignments: List[str] = list(specs.keys())

    try:
        prepare_student(student,
                        stogit_url,
                        do_clean=clean,
                        do_clone=not global_vars.CI and not skip_repo_update,
                        do_pull=not global_vars.CI and not skip_repo_update,
                        do_checkout=not global_vars.CI,
                        date=date)

        student_result = StudentResult(name=student)

        if record:
            record_student(student=student_result, specs=specs, assignments=assignments, basedir=basedir,
                           interact=interact, skip_web_compile=skip_web_compile)

        if analyze:
            analyze_student(student=student_result, specs=specs, check_for_branches=not skip_branch_check)

        if student_result.unmerged_branches:
            for result in student_result.results:
                result.warnings.unmerged_branches = student_result.unmerged_branches

        if date:
            reset(student)

        return student_result

    except Exception as err:
        if global_vars.DEBUG:
            raise err
        else:
            return StudentResult(name=student, error=str(err))


def prepare_student(student: str,
                    stogit_url: str,
                    do_clean: bool,
                    do_clone: bool,
                    do_pull: bool,
                    do_checkout: bool,
                    date: str = ''):
    if do_clean:
        remove(student)
    if do_clone:
        clone_student(student, base_url=stogit_url)
    if do_pull:
        stash(student)
        pull(student)
    if do_checkout:
        checkout_date(student, date=date)


def prepare_student_repo(student: str,
                         stogit_url: str,
                         do_clean: bool,
                         do_clone: bool,
                         do_pull: bool,
                         do_checkout: bool,
                         date: str = '') -> str:
    prepare_student(student,
                    stogit_url,
                    do_clean,
                    do_clone,
                    do_pull,
                    do_checkout,
                    date)

    return student



from inspect import isgenerator
from stograde.common.flatten import flatten


def test_flatten():
    assert isgenerator(flatten([]))
    assert list(flatten([1])) != [[1]]

    assert list(flatten([[]])) == []
    assert list(flatten([1, [2], [3]])) == [1, 2, 3]
    assert list(flatten([1, [2, 3], [[4]]])) == [1, 2, 3, 4]



import logging
import re
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from ..student.student_result import StudentResult

LAB_REGEX = re.compile(r'^LAB', re.IGNORECASE)


def ci_analyze(student_results: List['StudentResult']) -> bool:
    passing = True
    for student_result in student_results:
        for result in student_result.results:
            try:
                for file in result.file_results:
                    # Alert student about any missing files
                    if file.file_missing and not file.optional:
                        logging.error("{}: File {} missing".format(result.spec_id, file.file_name))
                        if not re.match(LAB_REGEX, result.spec_id):
                            passing = False
                    else:
                        # Alert student about any compilation errors
                        for compilation in file.compile_results:
                            if compilation.status != 'success':
                                if file.compile_optional:
                                    logging.warning("{}: File {} compile error (This did not fail the build)"
                                                    .format(result.spec_id, file.file_name))
                                else:
                                    logging.error("{}: File {} compile error:\n\n\t{}"
                                                  .format(result.spec_id, file.file_name,
                                                          compilation.output.replace("\n", "\n\t")))
                                    if not re.match(LAB_REGEX, result.spec_id):
                                        passing = False
            except KeyError:
                logging.error("KeyError with {}".format(result.spec_id))
    return passing



import copy
import io
import os
import pty
import subprocess
from typing import List, Optional, Tuple

from ..common.run_status import RunStatus


def run(cmd: List[str],
        *,
        interact: bool = False,
        input_data: Optional[bytes] = None,
        timeout: Optional[float] = None) -> Tuple[RunStatus, str, bool]:
    if interact:
        return run_interactive(cmd)
    else:
        return run_static(cmd, input_data, timeout)


def run_interactive(cmd: List[str]) -> Tuple[RunStatus, str, bool]:
    status = RunStatus.SUCCESS
    result = None

    print('Recording {}. Send EOF (^D) to end.'.format(cmd), end='\n\n')

    # This is mostly taken from the stdlib's `pty` docs
    with io.BytesIO() as script:
        def read(fd):
            data = os.read(fd, 1024)
            script.write(data)
            return data

        # TODO: update this with try/except clauses as we find exceptions
        pty.spawn(cmd, read)

        try:
            result = script.getvalue().decode(encoding='utf-8')
        except UnicodeDecodeError:
            result = script.getvalue().decode(encoding='cp437')

    print('\nSubmission recording completed.')

    runagain = input('Do you want to run the submission again? [y/N]: ')
    again = runagain.lower().startswith('y')

    return status, result, again


def run_static(cmd: List[str],
               input_data: Optional[bytes] = None,
               timeout: Optional[int] = None) -> Tuple[RunStatus, str, bool]:
    status = RunStatus.SUCCESS
    result = ''

    try:
        proc_result = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            timeout=timeout,
            input=input_data,
            env=copy_env(),
            check=True)

        if hasattr(proc_result, 'stdout'):
            result = proc_result.stdout

        # if hasattr(proc_result, 'stderr'):
        #     if str(proc_result.stderr)

    except subprocess.CalledProcessError as err:
        status = RunStatus.CALLED_PROCESS_ERROR
        result = err.output if err.output else str(err)

    except subprocess.TimeoutExpired as err:
        status = RunStatus.TIMEOUT_EXPIRED
        result = err.output if err.output else str(err)

    except FileNotFoundError as err:
        status = RunStatus.FILE_NOT_FOUND
        result = str(err)

    except ProcessLookupError as err:
        try:
            status, result, _ = run(cmd, input_data=input_data, timeout=timeout)
        except:
            status = RunStatus.PROCESS_LOOKUP_ERROR
            result = str(err)

    try:
        if not isinstance(result, str):
            result = str(result, 'utf-8')
    except UnicodeDecodeError:
        result = str(result, 'cp437')

    return status, result, False


# This is to catch glibc errors, because it prints to /dev/tty
# instead of stderr. See https://stackoverflow.com/a/27797579
def copy_env():
    env = copy.copy(os.environ)
    env["LIBC_FATAL_STDERR_"] = "1"
    return env



import functools
import os
from PyInquirer import prompt, style_from_dict, Token
from typing import List, TYPE_CHECKING

from . import server
from ..common import chdir
from ..process_assignment import import_supporting, remove_supporting
from ..process_file import process_file
from ..process_file.file_result import FileResult
from ..process_file.process_file import get_file
from ..student.process_student import prepare_student_repo
from ..toolkit.process_repos import process_parallel_repos

if TYPE_CHECKING:
    from ..specs.spec import Spec


def launch_cli(base_dir: str,
               clean: bool,
               date: str,
               no_progress_bar: bool,
               skip_repo_update: bool,
               spec: 'Spec',
               stogit_url: str,
               students: List[str],
               workers: int):
    """Start the web grading CLI"""
    usernames = [
        '{} NO SUBMISSION'.format(student)
        if not os.path.exists('{}/{}'.format(student, spec.id))
        else student
        for student in students
    ]

    print('Loading repos. Please wait...')

    with chdir(os.path.join(base_dir, 'students')):
        single_repo = functools.partial(
            prepare_student_repo,
            stogit_url=stogit_url,
            do_clean=clean,
            do_clone=not skip_repo_update,
            do_pull=not skip_repo_update,
            do_checkout=True,
            date=date)

        process_parallel_repos(students=students,
                               no_progress_bar=no_progress_bar,
                               workers=workers,
                               operation=single_repo)

    while True:
        student = ask_student(usernames)

        if not student or student == 'QUIT':
            return
        elif 'NO SUBMISSION' in student:
            continue

        files = check_student(student, spec, base_dir)

        if files:
            ask_file(files, student, spec, base_dir)


def ask_student(usernames: List[str]) -> str:
    """Ask user to select a student"""
    style = style_from_dict({
        Token.QuestionMark: '#959ee7 bold',
        Token.Selected: '#959ee7',
        Token.Pointer: '#959ee7 bold',
        Token.Answer: '#959ee7 bold',
    })
    questions = [
        {
            'type': 'list',
            'name': 'student',
            'message': 'Choose student',
            'choices': ['QUIT', *usernames]
        }
    ]

    student = prompt(questions, style=style)

    if not student:
        return ''

    return student['student']


def check_student(student: str,
                  spec: 'Spec',
                  base_dir: str):
    """Process student's files and populate file list"""
    files = []
    if os.path.exists('{}/{}'.format(student, spec.id)):
        print('Processing...')
        with chdir('{}/{}'.format(student, spec.id)):
            # prepare the current folder
            supporting_dir, written_files = import_supporting(spec=spec,
                                                              basedir=base_dir)

            for file in spec.files:
                if file.options.web_file:
                    result = FileResult(file_name=file.file_name)
                    exists = get_file(file_spec=file,
                                      file_result=result)

                    description = file.file_name

                    if not exists:
                        if result.optional:
                            description = '{} MISSING (OPTIONAL)'.format(file.file_name)
                        else:
                            description = '{} MISSING'.format(file.file_name)

                    files = files + [description]
                else:
                    continue
            # and we remove any supporting files
            remove_supporting(written_files)

    return files


def ask_file(files: List[str],
             student: str,
             spec: 'Spec',
             basedir: str):
    """Ask user to select a file to view"""
    style = style_from_dict({
        Token.QuestionMark: '#e3bd27 bold',
        Token.Selected: '#e3bd27',
        Token.Pointer: '#e3bd27 bold',
        Token.Answer: '#e3bd27 bold',
    })

    while True:
        questions = [
            {
                'type': 'list',
                'name': 'file',
                'message': 'Choose file',
                'choices': ['BACK'] + files,
            }
        ]
        file = prompt(questions, style=style)

        # File has been selected so process and display it
        if file and file['file'] != 'BACK':
            file_spec = None
            for f in spec.files:
                if f.file_name == file['file']:
                    file_spec = f
                    break
            if file_spec:
                with chdir('{}/{}'.format(student, spec.id)):
                    # prepare the current folder
                    supporting_dir, written_files = import_supporting(spec=spec,
                                                                      basedir=basedir)
                    process_file(file_spec=file_spec,
                                 cwd=os.getcwd(),
                                 supporting_dir=supporting_dir,
                                 interact=False,
                                 skip_web_compile=False)

                    server.work_dir = os.getcwd()

                    # and we remove any supporting files
                    remove_supporting(written_files)

        else:
            return


def is_web_spec(spec: 'Spec') -> bool:
    """Check if the spec contains any web files"""
    web_spec = False
    for file in spec.files:
        if file.options.web_file:
            web_spec = True
            break
    return web_spec



import functools
import logging
from concurrent.futures import as_completed, ProcessPoolExecutor
from typing import Dict, List

from ..common import chdir
from ..specs.spec import Spec
from ..student import process_student
from ..student.student_result import StudentResult
from ..toolkit.progress_bar import make_progress_bar


def process_students(specs: Dict[str, 'Spec'],
                     students: List[str],
                     *,
                     analyze: bool,
                     base_dir: str,
                     clean: bool,
                     date: str,
                     interact: bool,
                     no_progress_bar: bool,
                     record: bool,
                     skip_branch_check: bool,
                     skip_repo_update: bool,
                     skip_web_compile: bool,
                     stogit_url: str,
                     workers: int,
                     work_dir: str) -> List['StudentResult']:
    with chdir(work_dir):
        single_analysis = functools.partial(
            process_student,
            analyze=analyze,
            basedir=base_dir,
            clean=clean,
            date=date,
            interact=interact,
            skip_branch_check=skip_branch_check,
            skip_repo_update=skip_repo_update,
            record=record,
            specs=specs,
            skip_web_compile=skip_web_compile,
            stogit_url=stogit_url
        )

        results: List['StudentResult'] = []

        if workers > 1:
            print_progress = make_progress_bar(students, no_progress_bar=no_progress_bar)
            with ProcessPoolExecutor(max_workers=workers) as pool:
                futures = [pool.submit(single_analysis, name) for name in students]
                for future in as_completed(futures):
                    result: 'StudentResult' = future.result()
                    print_progress(result.name)
                    results.append(result)
        else:
            for student in students:
                logging.debug('Processing {}'.format(student))
                result: 'StudentResult' = single_analysis(student)
                results.append(result)

    return results



from dataclasses import dataclass, field
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from ..process_file.compile_result import CompileResult
    from ..process_file.test_result import TestResult


@dataclass
class FileResult:
    """The results from compiling and testing an assignment file"""
    file_name: str  # Name of the file
    contents: str = ''  # Contents of the file
    compile_results: List['CompileResult'] = field(default_factory=list)  # Results of each compilation
    test_results: List['TestResult'] = field(default_factory=list)  # Results of each test
    file_missing: bool = False  # Is the file missing
    last_modified: str = ''  # Last modification date according to git
    other_files: str = ''  # Other files in the directory (used if file is missing)
    optional: bool = False  # Is the file not required to exist
    compile_optional: bool = False  # Is the file not required to compile



from dataclasses import dataclass

from ..common.run_status import RunStatus


@dataclass
class TestResult:
    """Result from testing an assignment file"""
    command: str  # Command used to test
    output: str  # Output from the test
    error: bool  # Did the command return an error code
    status: RunStatus  # Status of the test
    truncated: bool = False  # Was the output truncated
    truncated_after: int = 0  # How much was it truncated



import logging
from os import path
from typing import Dict, List, TYPE_CHECKING

from ..common import chdir
from ..process_assignment.process_assignment import process_assignment
from ..process_assignment.record_result import RecordResult
from ..process_assignment.warning_unmerged_branches import find_unmerged_branches
from ..toolkit import global_vars

if TYPE_CHECKING:
    from ..specs.spec import Spec
    from ..student.student_result import StudentResult


def record_student(*,
                   student: 'StudentResult',
                   specs: Dict[str, 'Spec'],
                   assignments: List[str],
                   basedir: str,
                   interact: bool,
                   skip_web_compile: bool):
    results = []
    if assignments:
        directory = student.name if not global_vars.CI else '.'
        with chdir(directory):
            find_unmerged_branches(student)

            for _, spec in specs.items():
                if spec.id in assignments:
                    logging.debug("Recording {}'s {}".format(student.name, spec.id))
                    if path.exists(spec.folder):
                        with chdir(spec.folder):
                            assignment_result = process_assignment(student=student,
                                                                   spec=spec,
                                                                   basedir=basedir,
                                                                   interact=interact,
                                                                   skip_web_compile=skip_web_compile)
                    else:
                        assignment_result = RecordResult(spec_id=spec.id,
                                                         student=student.name)
                        assignment_result.warnings.assignment_missing = True

                    results.append(assignment_result)

    student.results = results



import datetime
import logging
import os.path
import sys
from os import getcwd
from typing import Dict, TYPE_CHECKING

from .args import process_args
from .check_dependencies import check_dependencies
from .find_update import update_available
from .process_repos import create_students_dir
from .stogit_url import compute_stogit_url
from ..specs import create_data_dir, filter_assignments, find_all_specs, load_specs

if TYPE_CHECKING:
    from ..specs.spec import Spec


def main():
    base_dir = getcwd()
    args, students, assignments = process_args()
    command: str = args['command']
    command_func = args['func']
    course: str = args['course']
    skip_dependency_check: bool = args['skip_dependency_check']
    skip_version_check: bool = args['skip_version_check']
    stogit: str = args['stogit']

    if not skip_version_check:
        current_version, new_version = update_available()
        if new_version:
            print(('v{} is available: you have v{}. '
                   'Try "pip3 install --no-cache --user --upgrade stograde" '
                   'to update.').format(new_version, current_version), file=sys.stderr)

    if not skip_dependency_check:
        check_dependencies()

    if not os.path.exists('data'):
        create_data_dir(course, base_dir)

    stogit_url = compute_stogit_url(stogit=stogit, course=course, _now=datetime.date.today())

    if not os.path.exists('students') and command != 'ci':
        create_students_dir(base_dir=base_dir)

    if command == 'repo':
        command_func(students=students,
                     stogit_url=stogit_url,
                     base_dir=base_dir,
                     no_progress_bar=args['no_progress_bar'],
                     workers=args['workers'])
        sys.exit(0)

    elif command == 'table':
        assignments = [path.split('/')[-1].split('.')[0]
                       for path in find_all_specs(os.path.join(base_dir, 'data', 'specs'))]

    date: str = args['date']
    skip_spec_update: bool = args['skip_spec_update']

    if date:
        logging.debug('Checking out {}'.format(date))

    assignments = filter_assignments(assignments)

    loaded_specs: Dict[str, 'Spec'] = load_specs(assignments,
                                                 data_dir=os.path.join(base_dir, 'data'),
                                                 skip_spec_update=skip_spec_update)

    if not loaded_specs:
        print('No specs loaded!')
        sys.exit(1)

    # Call function to handle SubCommand
    command_func(specs=loaded_specs,
                 students=students,
                 base_dir=base_dir,
                 stogit_url=stogit_url,
                 args=args)



from shutil import get_terminal_size
import sys
from typing import List

CHAR = '·' if sys.stderr.encoding == 'UTF-8' else '='


def progress_bar(size, current, message=''):
    cols, _ = get_terminal_size()

    filled = [CHAR for i in range(current)]
    empty = [' ' for i in range(size - current)]
    bar = ''.join(filled + empty)

    line = '[{}] {}'.format(bar, message)
    spacers = ' ' * (cols - len(line))

    result = line + spacers
    result = result[:cols]
    print('\r' + result, end='', file=sys.stderr)


def make_progress_bar(students: List[str], no_progress_bar: bool = False):
    if no_progress_bar:
        return lambda _: None

    size = len(students)
    remaining = set(students)
    invocation_count = 0

    def increment(username):
        nonlocal remaining
        nonlocal invocation_count
        remaining.remove(username)
        invocation_count += 1
        msg = ', '.join(sorted(remaining))
        progress_bar(size, invocation_count, message=msg)

    msg = ', '.join(sorted(remaining))
    progress_bar(size, invocation_count, message=msg)
    return increment



from enum import auto, Enum


class AssignmentStatus(Enum):
    """The completeness of the assignment"""
    SUCCESS = auto()  # All required files are present
    PARTIAL = auto()  # Some files are missing
    MISSING = auto()  # No files are present



"""
Adapted from bridge.py from the SD course.

Simple HTTP server in python, that invokes an executable to handle requests.
Send a POST request::
    curl -d "3ok" http://localhost:25199
"""
from http.server import BaseHTTPRequestHandler, HTTPServer
import logging
import os
from subprocess import *
import sys

from ..common import chdir

exe_name = "a.out"
work_dir = "."
exe_mtime = -1
message_id = {}
yaml_files_name = "app.files"
mem_file_name = "app.mem"
yaml_mtime = -1
separator = '\n---\n'


def has_top_key(key, yaml):
    return yaml.startswith(key) or -1 != yaml.find('\n' + key)


def get_top_key(key, yaml):
    if yaml.startswith(key):
        res = yaml[len(key):]
    else:
        pos = yaml.find('\n' + key)
        if -1 == pos:
            return False
        res = yaml[pos + len(key) + 1:]
    # trim off after newline
    pos2 = res.find('\n')
    if -1 == pos2:
        return res
    res = res[:pos2]
    return res


def load_binary(filename):
    with open(filename, 'rb') as fh:
        return fh.read()


class S(BaseHTTPRequestHandler):
    def _set_headers(self, content):
        self.send_response(200)
        self.send_header('Content-type', content)
        self.send_header('Access-Control-Allow-Origin', '*')
        self.end_headers()

    def use_exe(self, first_name, yaml_part, incoming_data):
        global exe_name, exe_mtime, message_id, separator
        new_mtime = os.path.getmtime(exe_name)
        is_poll = False
        just_recompiled = False
        if new_mtime > exe_mtime:  # restart!
            message_id[first_name] = 0
            incoming_data = "- message_id: " + str(message_id[first_name]) + "\n"
            incoming_data += "- first_name: " + first_name + separator
            incoming_data = incoming_data.encode()
            logging.debug("recompiled")
            just_recompiled = True
        else:
            if has_top_key('- initialize: ', yaml_part):  # initialization
                message_id[first_name] = 0
                incoming_data = "- message_id: " + str(message_id[first_name]) + "\n"
                incoming_data += "- first_name: " + first_name + separator
                incoming_data = incoming_data.encode()
                logging.debug('init_client')
            elif has_top_key('- event_info:', yaml_part):  # user interaction
                message_id[first_name] += 1
                incoming_data = ("- message_id: " + str(message_id[first_name]) + "\n").encode() + incoming_data
                logging.debug('event')
            else:  # just a poll
                # is_poll = True
                incoming_data = ("- message_id: " + str(message_id[first_name]) + " poll\n").encode() + incoming_data
        if is_poll:
            outgoing_data = 'same_exe'
            outgoing_data = bytes(outgoing_data, "utf-8")
        else:
            vverbose = False
            if vverbose:
                logging.debug(" running with incoming len:\n")
                sys.stderr.buffer.write(incoming_data)
                logging.debug("\n")
            try:
                proc = Popen(os.path.join(os.getcwd(), exe_name), stdin=PIPE, stdout=PIPE, stderr=PIPE)
                try:
                    (stdout_, stderr_) = proc.communicate(incoming_data, timeout=2)
                except TimeoutExpired:
                    proc.kill()
                    (stdout_, stderr_) = proc.communicate()
                    stderr_ = b'timeout expired:  your code may have an infinite loop!\n' + stderr_
            except FileNotFoundError:
                stdout_ = ""
                stderr_ = "File {} not found".format(os.path.join(os.getcwd(), exe_name))
            if 'str' == type(stdout_):
                stdout_ = bytes(stdout_, 'utf-8')
                stderr_ = bytes(stderr_, 'utf-8')
            if vverbose:
                logging.debug("\n STDOUT ")
                sys.stderr.buffer.write(stdout_)
                logging.debug("\n")
            keyword = b'React Native'
            if stdout_.startswith(keyword):
                if vverbose:
                    logging.debug(" keyword ")
                stdout_ = stdout_[len(keyword):]
            elif b'same_exe' == stdout_:
                if vverbose:
                    logging.debug(" was_poll ")
            elif b'timeout' == stdout_:
                if vverbose:
                    logging.debug(" timeout ")
            else:
                if vverbose:
                    logging.debug(" no_keyword ")
                if vverbose:
                    sys.stderr.buffer.write(stdout_)
                if os.path.isfile(yaml_files_name):
                    if just_recompiled:
                        logging.debug(" sending static ")
                    stdout_ = self.use_static_yaml_inner()
                else:
                    stdout_ = b''
            outgoing_data = stdout_
            if len(stderr_) > 0:
                outgoing_data = b'- stderr: ' + stderr_ + bytes(chr(0), "utf-8") + outgoing_data
        self._set_headers('text/html')
        self.wfile.write(outgoing_data)
        exe_mtime = new_mtime

    def use_static_yaml_inner(self):
        global separator, yaml_mtime
        new_mtime = os.path.getmtime(yaml_files_name)
        new_mtime2 = os.path.getmtime(mem_file_name)
        if new_mtime2 > new_mtime:
            new_mtime = new_mtime2
        if new_mtime > yaml_mtime:
            logging.debug("sending static yaml")
        outgoing_data = ''
        with open(yaml_files_name, 'r') as fh:
            for line in fh:
                line = line.lstrip()
                if line[0] != '#':
                    line = line.rstrip()
                    with open(line, 'r') as fh2:
                        outgoing_data += '\n' + fh2.read()
            outgoing_data += separator
            mem_data = b''
            with open(mem_file_name, 'rb') as fh3:
                mem_data += fh3.read()
            while len(mem_data) < 10000:  # fill in the rest with null bytes
                mem_data += b'\0'
            outgoing_data = outgoing_data.encode('utf8') + mem_data
            # print('outgoing_data is as follows:\n', outgoing_data)
            yaml_mtime = new_mtime
        return outgoing_data

    def use_static_yaml(self):
        outgoing_data = self.use_static_yaml_inner()
        self._set_headers('text/html')
        self.wfile.write(outgoing_data)

    def do_GET(self):
        if self.path.endswith(".png"):
            self._set_headers('image/png')
            self.wfile.write(load_binary(self.path[self.path.rfind("/") + 1:]))

    def do_POST(self):
        with chdir(work_dir):
            content_length = int(self.headers['Content-Length'])
            incoming_data = self.rfile.read(content_length)
            # parse the incoming data, to look for some key parts
            global exe_name, separator
            sep_pos = incoming_data.find(separator.encode())
            if -1 == sep_pos:
                yaml_part = ''
            else:
                yaml_part = incoming_data[:sep_pos].decode() + '\n'
            first_name = get_top_key('- first_name: ', yaml_part)

            # does the executable exist?
            if os.path.isfile(exe_name):
                logging.debug(" has_exe ")
                self.use_exe(first_name, yaml_part, incoming_data)
            elif os.path.isfile(yaml_files_name):
                logging.debug(" no_exe ")
                self.use_static_yaml()
            else:
                logging.debug(" nothing ")

    def log_message(self, format, *args):
        pass


def run_server(port=25199):
    server_address = ('', port)
    httpd = HTTPServer(server_address, S)
    httpd.serve_forever()



from tempfile import gettempdir
import logging
import sys
import os

from ..common import chdir
from ..specs import load_some_specs
from .args import process_args
from .process_student import process_student
from .parse_commits import parse_commits_for_assignments
from .emailify import emailify
from .send_email import send_email
from collections import Counter


def parse_gitlab6_webhook(payload):
    emails = [c['author']['email'] for c in payload['commits']]
    most_common_email = Counter(emails).most_common(1)[0][0]

    return {
        'name': payload['user_name'],
        'email': most_common_email,
        'branch': payload['ref'],
        'repo': payload['repository']['url'],
        'commits': payload['commits'],
        'repo_folder': str(payload['user_id']),
    }


def parse_gitlab9_webhook(payload):
    if payload['object_kind'] != 'push':
        raise Exception('Not a push event!')

    return {
        'name': payload['user_name'],
        'email': payload['user_email'],
        'branch': payload['ref'],
        'repo': payload['project']['git_ssh_url'],
        'commits': payload['commits'],
        'repo_folder': payload['project']['path_with_namespace'].split('/')[-1],
    }


def main():
    args = process_args()
    basedir = os.getcwd()

    logging.basicConfig(level=logging.DEBUG if args['debug'] else logging.WARNING)

    payload = args['data']

    print(payload)

    parsed_payload = {}
    try:
        parsed_payload = parse_gitlab6_webhook(payload)
    except KeyError:
        parsed_payload = parse_gitlab9_webhook(payload)

    print(parsed_payload)

    name = parsed_payload['name']
    email = parsed_payload['email']
    branch = parsed_payload['branch']
    repo = parsed_payload['repo']
    commits = parsed_payload['commits']
    repo_folder = parsed_payload['repo_folder']

    print('processing {}#{}'.format(repo, branch))
    print('author: {} <{}>'.format(name, email))
    print('destination: {}'.format(repo_folder))

    print('before push', payload['before'])
    print('after push', payload['after'])

    affected_assignments = parse_commits_for_assignments(commits)

    stringified_assignments = [''.join(pair) for pair in affected_assignments]
    print(stringified_assignments)

    specs = load_some_specs(stringified_assignments, basedir='./data')
    if not specs:
        print('no specs loaded!')
        sys.exit(1)

    # ensure that two runs of referee with the same repo don't interfere with each other
    with chdir(gettempdir()):
        # print('working in', os.getcwd())
        results, recordings = process_student(repo=repo,
                                              branch=branch,
                                              assignments=stringified_assignments,
                                              folder=repo_folder,
                                              specs=specs,
                                              basedir=basedir,
                                              debug=args['debug'])

    print('processing complete')

    email_blob = emailify(recordings=recordings,
                          name=name,
                          to=email,
                          debug=args['debug'])

    if args['send']:
        send_email(email_blob)
        print('email sent')
    else:
        print('Not sending email: no --send flag')
        print()
        print(email_blob)



from textwrap import dedent
from pyfakefs import fake_filesystem
from collections import Mapping
from stograde.toolkit.get_students import get_students


def write_students_file(fs, students):
    fs.create_file('./students.txt')
    with open('./students.txt', 'w', encoding='utf-8') as outfile:
        outfile.write(dedent(students).strip())


def test_get_students_single_name(fs):
    write_students_file(fs, 'rives')

    # verify that we got a dict
    assert isinstance(get_students(), Mapping)

    # this dict should have a 'my' section
    assert 'my' in get_students()

    # and the "my" section should include one username
    assert get_students()['my'] == ['rives']


def test_get_students_multiple_names(fs):
    write_students_file(fs, '''
    rives
    piersonv
    magnusow
    ''')

    # verify that we got a dict
    assert isinstance(get_students(), Mapping)

    # this dict should have a 'my' section
    assert 'my' in get_students()

    # and the "my" section should include one username
    assert get_students()['my'] == ['rives', 'piersonv', 'magnusow']


def test_get_students_single_section(fs):
    write_students_file(fs, '''
    [alt]
    rye
    ''')

    assert 'alt' in get_students()
    assert get_students()['alt'] == ['rye']


def test_get_students_multiple_sections(fs):
    write_students_file(fs, '''
    [my]
    rives
    piersonv
    magnusow

    [section-a]
    piersonv

    [section-b]
    magnusow
    ''')

    assert 'my' in get_students()
    assert 'section-a' in get_students()
    assert 'section-b' in get_students()

    assert get_students()['my'] == ['rives', 'piersonv', 'magnusow']
    assert get_students()['section-a'] == ['piersonv']
    assert get_students()['section-b'] == ['magnusow']


def test_get_students_missing_file(fs):
    os_module = fake_filesystem.FakeOsModule(fs)
    assert os_module.path.exists('./students.txt') == False
    assert get_students() == {'my': []}



import os
from typing import List, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from ..specs.spec import Spec


def import_supporting(*, spec: 'Spec', basedir: str) -> Tuple[str, List[str]]:
    """Copy supporting and input files into student's homework directory"""
    cwd = os.getcwd()
    supporting_dir = os.path.join(basedir, 'data', 'supporting')
    written_files = []

    # write the supporting files into the folder
    for file in spec.supporting_files:
        with open(os.path.join(supporting_dir, spec.id, file.file_name), 'rb') as infile:
            contents = infile.read()
        with open(os.path.join(cwd, file.destination), 'wb') as outfile:
            outfile.write(contents)
            written_files.append(file.destination)

    return supporting_dir, written_files


def remove_supporting(written_files: List[str]):
    """Remove supporting and input files after testing is complete"""
    try:
        for supporting_file in written_files:
            os.remove(supporting_file)
    except FileNotFoundError:
        pass




"""Given a spec, assuming we're in the homework folder, run the spec against the folder"""

import os
from typing import TYPE_CHECKING

from .record_result import RecordResult
from .submission_warnings import SubmissionWarnings
from .supporting import import_supporting, remove_supporting
from ..common import check_dates
from ..process_file import process_file
from ..toolkit import global_vars

if TYPE_CHECKING:
    from ..specs.spec import Spec
    from ..student.student_result import StudentResult


def process_assignment(*,
                       student: 'StudentResult',
                       spec: 'Spec',
                       basedir: str,
                       interact: bool,
                       skip_web_compile: bool) -> RecordResult:
    """Run a spec against the current folder"""
    cwd = os.getcwd()
    try:
        first_submit = ''

        if not global_vars.CI:
            first_date = check_dates(spec, cwd)
            first_submit = "First Submission for {}: {}".format(spec.id, first_date)

        result = RecordResult(spec_id=spec.id,
                              first_submission=first_submit,
                              student=student.name)

        # prepare the current folder
        supporting_dir, written_files = import_supporting(spec=spec,
                                                          basedir=basedir)
        for file_spec in spec.files:
            file_result = process_file(file_spec=file_spec,
                                       cwd=cwd,
                                       supporting_dir=supporting_dir,
                                       interact=interact,
                                       skip_web_compile=skip_web_compile)
            result.file_results.append(file_result)

        # now we remove any compiled binaries
        remove_execs(spec)

        # and we remove any supporting files
        remove_supporting(written_files)

        return result

    except Exception as err:
        if global_vars.DEBUG:
            raise err
        else:
            return RecordResult(student=student.name,
                                spec_id=spec.id,
                                warnings=SubmissionWarnings(recording_err=str(err)))


def remove_execs(spec: 'Spec'):
    """Remove executable files (identified by a .exec extension)"""
    try:
        for file in spec.files:
            os.remove('{}.exec'.format(file.file_name))
    except FileNotFoundError:
        pass



from typing import Tuple

from .run_status import RunStatus


def cat(filename: str) -> Tuple[RunStatus, str]:
    """Return the contents of a file. Replaces the `cat` command.

    This function took about ~148 time units per call, while
    run(['cat']) needed ~4688 time units.
    """
    try:
        with open(filename, 'r', encoding='utf-8') as infile:
            return RunStatus.SUCCESS, infile.read()
    except FileNotFoundError:
        return RunStatus.FILE_NOT_FOUND, ''
    except:
        return RunStatus.FAILURE, ''



from textwrap import dedent

from stograde.process_assignment.assignment_status import AssignmentStatus
from stograde.student.student_result import StudentResult
from stograde.toolkit.tabulate import find_columns, pad, MISSING, concat, symbol, columnize, get_nums, sort_by_hw_count, \
    sort_by_username, tabulate


def test_pad():
    assert pad("HW2", 12345) == "HW2  ", "should pad to the left to the width of the number"
    assert pad("HW2", 1) == "HW2", "should not trim the string if the number is shorter"
    assert pad(MISSING, 12345) == (MISSING * 5), "should use MISSING if given"


def test_find_columns():
    assert find_columns(10) == "1 2 3 4 5 6 7 8 9 10", "should return incremental numbers between 1 and the argument"


def test_symbol():
    assert symbol((5, AssignmentStatus.SUCCESS)) == '5', "return the number for successful assignments"
    assert symbol((5, AssignmentStatus.MISSING)) == MISSING, "returns MISSING for missing assignments"
    assert symbol((5, AssignmentStatus.PARTIAL), highlight_partials=False) == '5', \
        "returns the raw number if stdout can't handle fanciness"


def test_concat():
    assert concat({
        'hw1': AssignmentStatus.SUCCESS,
        'hw2': AssignmentStatus.SUCCESS,
    }, 2) == "1 2", "should turn an assignment list into a string"

    assert concat({
        'hw1': AssignmentStatus.SUCCESS,
        'hw2': AssignmentStatus.SUCCESS,
    }, 6) == "1 2 - - - -", "should output hyphens when no assignments are given"

    assert concat({
        'hw1': AssignmentStatus.SUCCESS,
        'hw6': AssignmentStatus.SUCCESS,
    }, 6) == "1 - - - - 6", "should handle missing assignments in between others"

    assert concat({
        'hw10': AssignmentStatus.SUCCESS,
        'hw11': AssignmentStatus.SUCCESS,
    }, 12) == "- - - - - - - - - 10 11 --", "should output two hyphens for two-digit numbers"


def test_columnize():
    student = StudentResult(name='rives',
                            homeworks={
                                'hw1': AssignmentStatus.SUCCESS,
                                'hw2': AssignmentStatus.SUCCESS,
                            },
                            labs={
                                'lab1': AssignmentStatus.SUCCESS,
                            })

    assert columnize(student, 'rives', 2, 1, 1) == "rives  | 1 2 | 1 | -"
    assert columnize(student, 'long_username', 2, 1, 1) == "rives          | 1 2 | 1 | -"
    assert columnize(student, 'rives', 5, 1, 1) == "rives  | 1 2 - - - | 1 | -"
    assert columnize(student, 'rives', 2, 5, 1) == "rives  | 1 2 | 1 - - - - | -"

    student2 = StudentResult(name='rives',
                             unmerged_branches=['branch'],
                             homeworks={
                                 'hw1': AssignmentStatus.SUCCESS,
                                 'hw2': AssignmentStatus.SUCCESS,
                             },
                             labs={
                                 'lab1': AssignmentStatus.SUCCESS,
                             })

    assert columnize(student2, 'rives', 2, 1, 1) == "\033[1mrives\033[0m  | 1 2 | 1 | -"

    student3 = StudentResult(name='rives',
                             error='an error occurred')

    assert columnize(student3, 'rives', 2, 1, 1) == "{username}  | {error}".format(username=student3.name,
                                                                                   error=student3.error)


def test_get_nums():
    assert get_nums([
        StudentResult(name='rives',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                      },
                      labs={
                          'lab1': AssignmentStatus.SUCCESS,
                      }),
    ]) == (2, 1, 0)

    assert get_nums([
        StudentResult(name='rives',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                      },
                      labs={
                          'lab1': AssignmentStatus.SUCCESS,
                      }),
        StudentResult(name='rives',
                      homeworks={
                          'hw8': AssignmentStatus.SUCCESS,
                          'hw9': AssignmentStatus.SUCCESS,
                      },
                      labs={
                          'lab10': AssignmentStatus.SUCCESS,
                      }),
    ]) == (9, 10, 0)

    assert get_nums([
        StudentResult(name='student',
                      homeworks={},
                      labs={}),
    ]) == (0, 0, 0)

    assert get_nums([
        StudentResult(name='student',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                      })
    ]) == (1, 0, 0)

    assert get_nums([
        StudentResult(name='student',
                      labs={
                          'lab1': AssignmentStatus.SUCCESS,
                      })
    ]) == (0, 1, 0)


def test_sort_by_hw_count():
    students = [
        StudentResult(name='rives1',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                          'hw3': AssignmentStatus.SUCCESS,
                          'hw4': AssignmentStatus.SUCCESS,
                      }),
        StudentResult(name='rives2',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                          'hw3': AssignmentStatus.SUCCESS,
                      }),
        StudentResult(name='rives3',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                      }),
    ]

    assert sorted(students, key=sort_by_hw_count) == list(reversed(students))


def test_sort_by_username():
    students = [
        StudentResult(name='rives1',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                          'hw3': AssignmentStatus.SUCCESS,
                          'hw4': AssignmentStatus.SUCCESS,
                      }),
        StudentResult(name='rives2',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                          'hw3': AssignmentStatus.SUCCESS,
                      }),
        StudentResult(name='rives3',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                      }),
    ]

    assert sorted(students, key=sort_by_username) == students


def test_tabulate():
    students = [
        StudentResult(name='rives3',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                          'hw3': AssignmentStatus.SUCCESS,
                          'hw4': AssignmentStatus.SUCCESS,
                      }),
        StudentResult(name='rives2',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                          'hw3': AssignmentStatus.SUCCESS,
                      }),
        StudentResult(name='rives1',
                      homeworks={
                          'hw1': AssignmentStatus.SUCCESS,
                          'hw2': AssignmentStatus.SUCCESS,
                      },
                      labs={
                          'lab1': AssignmentStatus.SUCCESS,
                          'lab2': AssignmentStatus.SUCCESS,
                      },
                      worksheets={
                          'ws1': AssignmentStatus.SUCCESS
                      }),
    ]

    assert tabulate(students) == dedent("""
    USER    | 1 2 3 4 | 1 2 | 1
    --------+---------+-----+--
    rives1  | 1 2 - - | 1 2 | 1
    rives2  | 1 2 3 - | - - | -
    rives3  | 1 2 3 4 | - - | -
    """).strip()

    assert tabulate(students, sort_by='count') == dedent("""
    USER    | 1 2 3 4 | 1 2 | 1
    --------+---------+-----+--
    rives1  | 1 2 - - | 1 2 | 1
    rives3  | 1 2 3 4 | - - | -
    rives2  | 1 2 3 - | - - | -
    """).strip()



import logging
import os
import sys
from glob import iglob
from typing import Dict, List, TYPE_CHECKING

from .stogradeignore import load_stogradeignore
from ..specs.util import check_architecture, check_spec_dependencies
from ..toolkit import global_vars

if TYPE_CHECKING:
    from stograde.specs.spec import Spec


def filter_assignments(assignments: List[str]) -> List[str]:
    """Removes any assignments ignored by a .stogradeignore file"""
    if not global_vars.CI:
        return assignments
    else:
        filtered_assignments = set(assignments)

        ignored_assignments = load_stogradeignore()

        for assignment in ignored_assignments:
            if assignment in filtered_assignments:
                logging.info('Skipping {}: ignored by stogradeignore'.format(assignment))

        filtered_assignments = filtered_assignments.difference(ignored_assignments)

        return list(filtered_assignments)


def get_spec_paths(wanted_specs: List[str], spec_dir: str) -> List[str]:
    """Removes any missing specs from the list and returns a list of the paths
    of the remaining specs"""
    all_spec_files = find_all_specs(spec_dir)
    loadable_spec_files = {path.split('/')[-1].split('.')[0]: path for path in list(all_spec_files)}
    specs_to_load = set(loadable_spec_files.keys()).intersection(wanted_specs)
    missing_spec_files = set(wanted_specs).difference(loadable_spec_files.keys())

    for spec in missing_spec_files:
        logging.warning("No spec for {}".format(spec))

    return list(loadable_spec_files[filename] for filename in specs_to_load)


def find_all_specs(spec_dir: str):
    """Get a list of all .yaml files in the specs directory"""
    return list(iglob(os.path.join(spec_dir, '*.yaml')))


def filter_loaded_specs(specs: Dict[str, 'Spec']) -> Dict[str, 'Spec']:
    """Filters the loaded specs based on properties such as required architecture"""
    remaining_specs: Dict[str, 'Spec'] = {}

    for spec_id in specs.keys():
        spec_to_use = specs[spec_id]
        try:
            check_spec_dependencies(spec_to_use)
            if not check_architecture(spec_to_use):
                continue
        except KeyError:
            # Prevent lab0 directory from causing an extraneous output
            if spec_to_use.id != 'lab0':
                print('Spec {} does not exist'.format(spec_to_use.id), file=sys.stderr)
            continue
        remaining_specs[spec_id] = spec_to_use

    return remaining_specs



import logging
import os
from typing import List, TYPE_CHECKING

from .gist import post_gist
from .tabulate import asciiify
from ..formatters import format_collected_data, markdown

if TYPE_CHECKING:
    from ..student.student_result import StudentResult


def record_recording_to_disk(results, file_identifier):
    results = sorted(results, key=lambda file: file['student'])
    results = [file['content'] for file in results]
    output = '\n'.join(results)
    try:
        os.makedirs('logs', exist_ok=True)
        with open('logs/log-{}.md'.format(file_identifier), 'w', encoding='utf-8') as outfile:
            outfile.write(output)
    except Exception as err:
        logging.warning('error! could not write recording:', err)


def send_recording_to_gist(table, results, assignment):
    """Publish a table/result pair to a private gist"""

    # the "-" at the front is so that github sees it first and names the gist
    # after the homework
    table_filename = '-cs251 report %s table.txt' % assignment
    files = {
        table_filename: {'content': table},
    }

    for file in results:
        filename = file['student'] + '.' + file['type']
        files[filename] = {
            'content': file['content'].strip()
        }

    return post_gist('log for ' + assignment, files)


def save_recordings(results: List['StudentResult'],
                    table: str,
                    debug: bool = False,
                    gist: bool = False):
    """Take the list of recordings, group by assignment, then save to disk"""

    result_dict = format_collected_data(results,
                                        group_by='assignment',
                                        formatter=markdown)

    for assignment, content in result_dict.items():
        logging.debug("Saving recording for {}".format(assignment))
        if gist:
            table = asciiify(table)
            url = send_recording_to_gist(table, content, assignment)
            print(assignment, 'results are available at', url)
        else:
            record_recording_to_disk(content, assignment)



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .assignment_status import AssignmentStatus


@dataclass
class AnalysisResult:
    """The result of analyzing an assignment for a student"""
    status: 'AssignmentStatus'  # The submission status of the assignment



import logging
import os
from typing import Dict, TYPE_CHECKING

from ..common import chdir
from ..process_assignment.assignment_status import AssignmentStatus
from ..process_assignment.assignment_type import AssignmentType, get_assignment_type
from ..process_assignment.warning_unmerged_branches import find_unmerged_branches
from ..specs import get_filenames
from ..toolkit import global_vars

if TYPE_CHECKING:
    from ..specs.spec import Spec
    from ..student.student_result import StudentResult


def analyze_student(student: 'StudentResult', specs: Dict[str, 'Spec'], check_for_branches: bool):
    logging.debug("Analyzing {}'s assignments".format(student.name))

    directory = student.name if not global_vars.CI else '.'
    analyses = {}
    with chdir(directory):
        if check_for_branches and not global_vars.CI:
            find_unmerged_branches(student)
        for _, spec in specs.items():
            analyses[spec.id] = analyze_assignment(spec)

    for name, analysis in analyses.items():
        a_type = get_assignment_type(name)
        if a_type is AssignmentType.HOMEWORK:
            student.homeworks[name] = analysis
        elif a_type is AssignmentType.LAB:
            student.labs[name] = analysis
        elif a_type is AssignmentType.WORKSHEET:
            student.worksheets[name] = analysis


def analyze_assignment(spec: 'Spec') -> AssignmentStatus:
    if not os.path.exists(spec.folder):
        return AssignmentStatus.MISSING

    with chdir(spec.folder):
        files_that_do_exist = set(os.listdir('.'))
        files_which_should_exist = set(get_filenames(spec))
        intersection_of = files_that_do_exist.intersection(files_which_should_exist)

        if intersection_of == files_which_should_exist:
            # if every file that should exist, does: we're good.
            return AssignmentStatus.SUCCESS
        elif intersection_of:
            # if some files that should exist, do: it's a partial assignment
            return AssignmentStatus.PARTIAL
        else:
            # otherwise, none of the required files are there
            return AssignmentStatus.MISSING



from stograde.common.cat import cat
from stograde.common.run_status import RunStatus


def test_cat(fs):
    filename = 'foo.txt'
    contents = 'insert a story here'
    fs.create_file(filename, contents=contents)

    result = cat(filename)
    assert result[0] == RunStatus.SUCCESS
    assert result[1] == contents


def test_cat_missing(fs):
    result = cat('file.txt')
    assert result[0] == RunStatus.FILE_NOT_FOUND
    assert result[1] == ''



import os
import logging
import sys
from typing import List, TYPE_CHECKING

from ..common.run import run
from ..toolkit import global_vars

if TYPE_CHECKING:
    from ..specs.spec import Spec


def get_filenames(spec: 'Spec') -> List[str]:
    """returns the list of files from an assignment spec"""
    return [file.file_name for file in spec.files if not file.options.optional]


def check_spec_dependencies(spec: 'Spec'):
    for filepath in spec.dependencies:
        try:
            os.stat(filepath)
        except FileNotFoundError:
            logging.warning('spec {}: required file "{}" could not be found'.format(spec.id, filepath))


def check_architecture(spec: 'Spec') -> bool:
    _, user_arch, _ = run(['uname', '-m'])
    user_arch = user_arch.rstrip()
    spec_arch = spec.architecture

    if spec_arch is None or spec_arch == user_arch:
        return True
    else:
        if global_vars.CI:
            logging.info('Skipping {}: wrong architecture'.format(spec.id))
        else:
            print('{} requires {} architecture. You have {}'
                  .format(spec.id, spec.architecture, user_arch),
                  file=sys.stderr)
        return False



import os
import sys
from typing import Dict, List, TYPE_CHECKING

from .filter_specs import filter_loaded_specs, get_spec_paths
from .spec import create_spec
from ..common import chdir
from ..common.run import run
from ..common.run_status import RunStatus

if TYPE_CHECKING:
    from .spec import Spec


def load_specs(wanted_specs: List[str], data_dir: str, skip_spec_update: bool = True) -> Dict[str, 'Spec']:
    """Load the desired specs from the specs/ directory, filtering out any that are missing

    data/ directory should exist by this point with a repository
    """
    if not skip_spec_update:
        check_for_spec_updates(data_dir)

    # the repo has a /specs directory
    spec_dir = os.path.join(data_dir, 'specs')

    specs_to_load = get_spec_paths(wanted_specs, spec_dir)

    loaded_specs = [create_spec(filename, spec_dir) for filename in specs_to_load]  # Create list of Specs
    loaded_specs = {spec.id: spec for spec in loaded_specs}  # Convert to dict (id to Spec)
    loaded_specs = filter_loaded_specs(loaded_specs)

    return loaded_specs


def check_for_spec_updates(data_dir: str):
    """Check if the specs have any updates using git fetch"""
    with chdir(data_dir):
        res, _, _ = run(['git', 'fetch', 'origin'])

        if res is not RunStatus.SUCCESS:
            print("Error fetching specs", file=sys.stderr)

        _, out, _ = run(['git', 'log', 'HEAD..origin/master'])

    if out != '':
        print("Spec updates found - Updating", file=sys.stderr)
        with chdir(data_dir):
            run(['git', 'pull', 'origin', 'master'])



from dateutil.parser import parse
import logging
import os
from typing import TYPE_CHECKING

from .run import run

if TYPE_CHECKING:
    from ..specs.spec import Spec


def check_dates(spec: 'Spec', cwd: str) -> str:
    """ Port of the CheckDates program from C++
    Finds the first submission date for an assignment
    by comparing first commits for all files in the spec
    and returning the earliest date
    """

    dates = []

    for file in spec.files:
        # Run a git log on each file with earliest commits listed first
        try:
            status, res, _ = run(['git', 'log', '--reverse', '--pretty=format:%ad', '--date=iso8601', '--',
                                  os.path.join(cwd, file.file_name)])
        except Exception as e:
            logging.debug("CHECK_DATES Exception: {}".format(e))
            return "ERROR"

        # If we didn't get an error and got an output, add date to array
        if status == 'success' and res:
            # Parse the first line
            dates.append(parse(res.splitlines()[0]))

    # Return earliest date as a string with the format mm/dd/yyyy hh:mm:ss
    if not dates:
        return "ERROR"

    return min(dates).strftime("%x %X")



import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 6):
    sys.exit("The toolkit requires Python 3.6 or greater.\nYou have {}".format(sys.version_info))

install_requires = [
    'PyYAML==5.*',
    'requests >= 2.20.*',
    'termcolor == 1.*',
    'natsort==7.0.*',
    'appdirs == 1.4.*',
    'python-dateutil == 2.7.*',
    'PyInquirer == 1.0.*',
    'bidict == 0.19.*',
]

if sys.version_info < (3, 7):
    install_requires.append('dataclasses >= 0.6')

setup(
    name='stograde',
    version='4.0.0',
    description='The StoGrade Toolkit',
    author='Hawken Rives',
    author_email='hawkrives@gmail.com',
    url='https://github.com/stodevx/stograde',
    license='MIT',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='stolaf course-tooling',
    install_requires=install_requires,
    tests_require=['tox'],
    packages=find_packages(exclude=['tests', 'docs']),
    # see http://python-packaging.readthedocs.io/en/latest/command-line-scripts.html
    entry_points={
        'console_scripts': [
            'stograde=stograde.toolkit.__main__:main',
            'referee=stograde.referee.__main__:main',
        ],
    },
)



"""Deal with argument parsing for the toolkit"""

import argparse
from glob import glob
import logging
from logging import warning, debug
from natsort import natsorted
import os
import re
import sys
from typing import Any, Dict, List, Tuple

from . import global_vars
from .get_students import get_students as load_students_from_file
from .subcommands import do_ci, do_clean, do_record, do_table, do_update, do_web
from ..common import flatten, version
from ..specs import get_supported_courses

ASSIGNMENT_REGEX = re.compile(r'^(HW|LAB|WS)', re.IGNORECASE)


def build_argparser():
    """Construct the argument list and parse the passed arguments"""
    parser = argparse.ArgumentParser(description='The core of the StoGrade toolkit')

    # Common arguments
    parser.add_argument('-v', '--version', action='store_true',
                        help='Print the version of the toolkit and exit')

    base_options = argparse.ArgumentParser(description='common options')
    base_options.add_argument('--skip-version-check', '-V', action='store_true',
                              default=os.getenv('STOGRADE_SKIP_VERSION_CHECK', False) is not False,
                              help='Skips the pypi update check')
    base_options.add_argument('--skip-dependency-check', action='store_true',
                              help='Skip checking for dependencies')
    base_options.add_argument('--debug', action='store_true',
                              help='Enable debugging mode (throw errors, implies -w1)')
    base_options.add_argument('--no-progress-bar', action='store_true',
                              help='Hide the progress bar')
    base_options.add_argument('--workers', '-w', type=int, default=os.cpu_count(), metavar='N',
                              help='The number of operations to perform in parallel')

    # Repository url modifiers
    repo_selection = argparse.ArgumentParser(add_help=False)
    repo_selection.add_argument('--course', default='', metavar='ID',
                                help='Which course to evaluate '
                                     '(this sets a default stogit url and downloads the correct specs). '
                                     'Can be {} or one of the previous with /f## or /s## (i.e. sd/s19)'
                                .format(get_supported_courses()))
    repo_selection.add_argument('--stogit', metavar='URL',
                                help='Use an alternate stogit base URL (eg, git@stogit.cs.stolaf.edu:sd/s17)')

    # Recording options
    record_options = argparse.ArgumentParser(add_help=False)
    record_options.add_argument('--clean', action='store_true',
                                help='Remove student folders and re-clone them')
    record_options.add_argument('--skip-repo-update', '-R', action='store_true',
                                help='Do not update the student folders when checking')
    record_options.add_argument('--skip-spec-update', '-S', action='store_true',
                                help='Skip checking for spec updates')
    record_options.add_argument('--date', action='store', metavar='GIT_DATE',
                                help=('Check out last submission on GIT_DATE (eg, "last week", "tea time", "2 hrs ago")'
                                      '(see `man git-rev-list`)'))

    # Student selection
    student_selection = argparse.ArgumentParser(add_help=False)
    selection_args = student_selection.add_argument_group('student selection')
    selection_args.add_argument('--students', '--student', action='append', nargs='+', metavar='USERNAME', default=[],
                                help='Only iterate over these students.')
    selection_args.add_argument('--section', action='append', dest='sections', nargs='+', metavar='SECTION', default=[],
                                help='Only check these sections: my, all, a, b, etc.')

    # Table Printout Parent Parser
    table_options = argparse.ArgumentParser(add_help=False)
    table_options_args = table_options.add_argument_group('table printout options')
    table_options_args.add_argument('--sort', dest='sort_by', action='store', default='name', type=str,
                                    choices=['name', 'count'],
                                    help='Sort the students table')
    table_options_args.add_argument('--no-partials', '-P', action='store_true',
                                    help="Don't highlight partial submissions")

    # SubParsers
    sub_parsers = parser.add_subparsers(dest='command')

    # CI SubParser
    parser_ci = sub_parsers.add_parser('ci', help="Check a single student's assignment as part of a CI job")
    parser_ci.set_defaults(func=do_ci)

    # Record SubParser
    parser_record = sub_parsers.add_parser('record', help="Record students' work",
                                           parents=[base_options, record_options, repo_selection, table_options,
                                                    student_selection],
                                           conflict_handler='resolve')
    parser_record.set_defaults(func=do_record)
    parser_record.add_argument('assignments', nargs='+', metavar='HW',
                               help='An assignment to process')
    parser_record.add_argument('--table', '-t', action='store_true',
                               help='Show the overview table after recording is complete')
    parser_record.add_argument('--gist', action='store_true',
                               help='Post overview table and student recordings as a private gist')
    parser_record.add_argument('--interact', action='store_true',
                               help="Interact with each student's submission individually")
    parser_record.add_argument('--skip-branch-check', '-B', action='store_true',
                               help='Do not check for unmerged branches')
    parser_record.add_argument('--skip-web-compile', action='store_true',
                               help='Skip compilation and testing of files marked with web: true')

    # Repo SubParser
    parser_repo = sub_parsers.add_parser('repo', help='Tools for cloning and updating student repositories',
                                         conflict_handler='resolve')
    repo_sub_parsers = parser_repo.add_subparsers()
    repo_sub_parsers.add_parser('clean', aliases=['reclone'], help='Remove and reclone student repositories',
                                parents=[base_options, repo_selection, student_selection],
                                conflict_handler='resolve').set_defaults(func=do_clean)
    repo_sub_parsers.add_parser('update', aliases=['clone'], help='Clone and/or update student repos',
                                parents=[base_options, repo_selection, student_selection],
                                conflict_handler='resolve').set_defaults(func=do_update)

    # Table SubParser
    parser_table = sub_parsers.add_parser('table', help='Print an table of the assignments submitted by students',
                                          parents=[base_options, record_options, repo_selection, table_options,
                                                   student_selection],
                                          conflict_handler='resolve')
    parser_table.set_defaults(func=do_table)

    # Web SubParser
    parser_web = sub_parsers.add_parser('web', help='Run the CLI for grading React App files',
                                        parents=[base_options, record_options, repo_selection, student_selection],
                                        conflict_handler='resolve')
    parser_web.set_defaults(func=do_web)
    parser_web.add_argument('assignments', nargs=1, metavar='HW',
                            help='An assignment to process')
    parser_web.add_argument('--port', type=int, required=True,
                            help='Set the port for the server to use')

    return parser


def get_students(args: Dict[str, Any]) -> List[str]:
    """Get students from the command line or the students.txt file.
    Anything on the command line will override using the file"""
    sections = args['sections']
    students = args['students']

    people = [student for group in students for student in group]
    if not people:
        _all_students = load_students_from_file()
        if sections:
            collected = []
            for section_name in sections:
                student_set = []
                prefixed = 'section-{}'.format(section_name)

                if section_name in _all_students:
                    student_set = _all_students[section_name]
                elif prefixed in _all_students:
                    student_set = _all_students[prefixed]
                else:
                    warning('Neither section [section-{0}] nor [{0}] could not be found in ./students.txt'
                            .format(section_name))

                collected.append(student_set)
            people = [student for group in collected for student in group]
        else:
            people = list(flatten([_all_students[section] for section in _all_students]))

    return sorted(set(people))


def get_ci_assignments() -> List[str]:
    """Find assignments in the student's repository during a CI job"""
    all_assignments: List[str] = []
    dirs = glob('hw*') + glob('lab*') + glob('ws*')
    for line in dirs:
        all_assignments.append(line.split('/')[-1])
    return natsorted(set(all_assignments))


def process_args() -> Tuple[Dict[str, Any], List[str], List[str]]:
    """Process the arguments and create usable data from them"""
    parser = build_argparser()
    args = vars(parser.parse_args())

    if args['version']:
        print('version', version)
        sys.exit(0)

    global_vars.DEBUG = args.get('debug', False)
    logging.basicConfig(level=logging.DEBUG if global_vars.DEBUG else logging.WARNING)

    command: str = args['command']

    # ci SubCommand
    if command == 'ci':
        assignments = get_ci_assignments()
        students = [os.environ['CI_PROJECT_NAME']]
        args['course'] = os.environ['CI_PROJECT_NAMESPACE']

    # record SubCommand
    elif command == 'record':
        assignments = natsorted(set(args['assignments']))
        if len(assignments) == 0:
            print('stograde record requires at least one assignment')
            sys.exit(1)

        students = get_students(args)

    # table SubCommand
    elif command == 'table':
        assignments = []  # Will be filled in later once we know that the data/specs directory exists
        students = get_students(args)

    # web SubCommand
    elif command == 'web':
        assignments = natsorted(set(args['assignments']))
        if len(assignments) != 1:
            print('stograde web can only be used with one assignment at a time')
            sys.exit(1)

        students = get_students(args)

    # repo SubCommand
    elif command == 'repo':
        assignments = []
        students = get_students(args)

    else:
        print('Sub-command must be specified')
        sys.exit(1)

    if not students:
        print('No students selected\n'
              'Is your students.txt missing?')
        sys.exit(1)

    debug_print_args(args)
    debug_print_students(students)
    debug_print_assignments(assignments)

    return args, students, assignments


def debug_print_args(args):
    debug("Command Line Arguments:")
    for arg, value in args.items():
        debug("{}: {}".format(arg, str(value)))


def debug_print_students(students):
    debug("Students:")
    debug_print_grid(students)


def debug_print_assignments(things):
    debug("Assignments:")
    debug_print_grid(things)


def debug_print_grid(items):
    line = ""
    for i, item in enumerate(items):
        line += item.ljust(10)
        if i % 5 == 4:
            debug(line)
            line = ""



from glob import glob
import shlex

from .flatten import flatten
from .run import run


def expand_chunk(command_chunk):
    """Take a chunk of a command and expand it, like a shell"""
    if '*' in command_chunk:
        return glob(command_chunk)
    return [command_chunk]


def process_chunk(command):
    """Takes one piece of a pipeline and formats it for run_command"""
    # decode('unicode_escape') de-escapes the backslash-escaped strings.
    # like, it turns the \n from "echo Hawken \n 26" into an actual newline,
    # like a shell would.
    cmd = bytes(command, 'utf-8').decode('unicode_escape')

    # shlex splits commands up like a shell does.
    # I'm not entirely sure how it differs from just split(' '),
    # but figured it wasn't a bad thing to use.
    cmds = shlex.split(cmd)

    return list(flatten([expand_chunk(c) for c in cmds]))


def pipe(cmd_string):
    cmds = cmd_string.split(' | ')

    input_for_cmd = None
    for cmd in cmds[:-1]:
        _, input_for_cmd, _ = run(process_chunk(cmd), input_data=input_for_cmd)
        input_for_cmd = input_for_cmd.encode('utf-8')

    final_cmd = process_chunk(cmds[-1])
    return final_cmd, input_for_cmd



from dataclasses import dataclass, field
import os
from typing import List, TYPE_CHECKING
import yaml

from .spec_file import create_spec_file
from .supporting_file import create_supporting_file

if TYPE_CHECKING:
    from .spec_file import SpecFile
    from .supporting_file import SupportingFile


@dataclass
class Spec:
    id: str
    folder: str
    architecture: str = None
    dependencies: List[str] = field(default_factory=list)
    files: List['SpecFile'] = field(default_factory=list)
    supporting_files: List['SupportingFile'] = field(default_factory=list)


def create_spec(yaml_path: str, basedir: str) -> Spec:
    with open(os.path.join(basedir, yaml_path), 'r', encoding='utf-8') as yaml_file:
        loaded_file = yaml.safe_load(yaml_file)

    assert 'assignment' in loaded_file

    # assignment id and folder
    new_spec = Spec(id=loaded_file['assignment'],
                    folder=loaded_file.get('folder', loaded_file['assignment']))

    # assignment files
    if loaded_file.get('files', None) is not None:
        for file in loaded_file['files']:
            file_spec = create_spec_file(file)

            if loaded_file.get('tests', None) is not None:
                file_spec.add_from_tests(loaded_file['tests'])

            new_spec.files.append(file_spec)

    # dependencies
    dependencies = loaded_file.get('dependencies', [])
    if isinstance(dependencies, str):
        dependencies = [dependencies]
    assert isinstance(dependencies, list)

    # supporting/input files
    if loaded_file.get('supporting', None) is not None:
        for s_file in loaded_file['supporting']:
            new_spec.supporting_files.append(create_supporting_file(s_file))

    if loaded_file.get('inputs', None) is not None:
        for i_file in loaded_file['inputs']:
            new_spec.supporting_files.append(create_supporting_file(i_file))

    return new_spec



import contextlib
import os


@contextlib.contextmanager
def chdir(path: str):
    """Create a `with` block for changing into a directory"""
    current_dir = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(current_dir)



from stograde.formatters import format_collected_data, html, markdown
from stograde.common import group_by
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText


def emailify(*, recordings, name, to, debug):
    fancy = format_collected_data(recordings,
                                  group_by='student',
                                  formatter=html,
                                  debug=debug)

    plaintext = format_collected_data(recordings,
                                      group_by='student',
                                      formatter=markdown,
                                      debug=debug)

    grouped = dict(group_by(recordings, lambda s: s['student']))

    print(recordings)
    print(plaintext)
    print(grouped)

    # `grouped` is a dictionary, from username to list of assignment results
    if not recordings:
        raise Exception('No students found to create an email!')

    fancy_body = '\n'.join([r['content'] for r in list(fancy.values())[0]])
    plaintext_body = '\n'.join([r['content'] for r in list(plaintext.values())[0]])
    print(fancy_body)
    print(plaintext_body)

    msg = MIMEMultipart('alternative')
    msg['to'] = '{} <{}>'.format(name, to)
    msg['from'] = 'cs251-tas@stolaf.edu'
    msg['subject'] = build_subject(list(fancy.values())[0])
    msg['reply-to'] = 'cs251-tas@stolaf.edu'

    msg.attach(MIMEText(plaintext_body, 'plain'))
    msg.attach(MIMEText(fancy_body, 'html'))

    return msg


def build_subject(results):
    return '[referee] Results for ' + ', '.join([r['assignment'] for r in results]) + ' submission'



from dataclasses import dataclass


@dataclass
class SupportingFile:
    file_name: str
    destination: str


def create_supporting_file(s_file_spec) -> SupportingFile:
    if isinstance(s_file_spec, dict):
        assert 'file' in s_file_spec
        file_name = s_file_spec['file']
        destination = s_file_spec.get('destination', file_name)
    elif isinstance(s_file_spec, list):
        file_name = s_file_spec[0]
        if len(s_file_spec) == 1:
            destination = s_file_spec[0]
        else:
            destination = s_file_spec[1]
    elif isinstance(s_file_spec, str):
        file_name = s_file_spec
        destination = s_file_spec
    else:
        raise TypeError('Cannot parse "inputs:" and/or "supporting:": incorrect data type for a file: {}'
                        .format(type(s_file_spec)))

    return SupportingFile(file_name=file_name,
                          destination=destination)



from .cat import cat
from .chdir import chdir
from .check_submit_date import check_dates
from .dirsize import dirsize
from .find_unmerged_branches_in_cwd import find_unmerged_branches_in_cwd
from .flatten import flatten
from .parse_commit_msg_for_assignments import parse_commit_msg_for_assignments
from .pipe import pipe
from .run import run
from .version import version



import re


def determine_assignment_type(string):
    if string in ['hw', 'homework', 'homeworks']:
        return 'hw'
    if string in ['lab']:
        return 'lab'
    return None


def parse_commit_msg_for_assignments(message):
    matches = [m.groups() for m in re.finditer(r'([a-z]+) ?(\d+)', message.lower())]
    results = [(determine_assignment_type(kind), num) for kind, num in matches]

    return [pair for pair in results if pair[0] is not None]



from stograde.common.parse_commit_msg_for_assignments import parse_commit_msg_for_assignments
sample = parse_commit_msg_for_assignments


def test_parse_commit_msg_for_assignments():
    assert sample('''[hw1] rename secondComent to secondComment''') == [('hw', '1')]
    assert sample('''Add cleaned-up StarTrekDVDDriver code for HW21''') == [('hw', '21')]
    assert sample('''add lab 5,6,7,10''') == [('lab', '5')]
    assert sample('''added comments to hw8''') == [('hw', '8')]
    assert sample('''added DVD.h and DVD.cpp to hw21''') == [('hw', '21')]
    assert sample('''Added lab4 Part B''') == [('lab', '4')]
    assert sample('''Added louder1 from Lab 3 Part D''') == [('lab', '3')]
    assert sample('''adjusted hw10''') == [('hw', '10')]
    assert sample('''adjusted hw6''') == [('hw', '6')]
    assert sample('''adjusted hw7''') == [('hw', '7')]
    assert sample('''Class implementations - Lab 8 part B: Shape, Rectangle, Circle''') == [('lab', '8')]
    assert sample('''commiting partial lab5 for now''') == [('lab', '5')]
    assert sample('''committing lab8 Circle Rectangle Shape''') == [('lab', '8')]
    assert sample('''Complete - Lab 8 part B: ShapeDriver1 ShapeDriver2''') == [('lab', '8')]
    assert sample('''Complete - Lab 8 part C: proto-draw with shape drawing''') == [('lab', '8')]
    assert sample('''Complete DVD class implementation - Lab 7 part A: DVD class''') == [('lab', '7')]
    assert sample('''complete Hw20 submit''') == [('hw', '20')]
    assert sample('''complete Lab 10 part A''') == [('lab', '10')]
    assert sample('''Complete Lab 10 part A: make test (gitignore ignored the *.out files)''') == [('lab', '10')]
    assert sample('''Complete Lab 10 part A: make tests''') == [('lab', '10')]
    assert sample('''Complete Lab 10 part B''') == [('lab', '10')]
    assert sample('''COmplete lab 10 part B''') == [('lab', '10')]
    assert sample('''Complete Lab 10 part B: OpenGL examples''') == [('lab', '10')]
    assert sample('''Complete Lab 10 part C: Doxygen specs''') == [('lab', '10')]
    assert sample('''complete lab 10 part C: Doxygen specs''') == [('lab', '10')]
    assert sample('''Complete Lab 11''') == [('lab', '11')]
    assert sample('''Complete Lab 3 Sound''') == [('lab', '3')]
    assert sample('''Complete Lab 8 part A, part B: Shape''') == [('lab', '8')]
    assert sample('''Complete Lab 8 part A: ColorPoint2.h''') == [('lab', '8')]
    assert sample('''complete Lab 9 part A:Shape.dia''') == [('lab', '9')]
    assert sample('''Complete Lab10 part B: openGL examples''') == [('lab', '10')]
    assert sample('''Complete Lab10 part C: DOxygen specs''') == [('lab', '10')]
    assert sample('''complete lab5 Part B''') == [('lab', '5')]
    assert sample('''complete Lab8 part A: colorpoint2.h''') == [('lab', '8')]
    assert sample('''Complete lab9 part A: lab9.dia''') == [('lab', '9'), ('lab', '9')]
    assert sample('''Complete Lab9 part A: laundry.dia''') == [('lab', '9')]
    assert sample('''error in part B of Lab 12, cann't get read in file to work.''') == [('lab', '12')]
    assert sample('''final submit hw18: corrected DVD.h''') == [('hw', '18')]
    assert sample('''first part of lab 7''') == [('lab', '7')]
    assert sample('''fix .gitignore to not ignore hw8/''') == [('hw', '8')]
    assert sample('''fix lab2 gitignore''') == [('lab', '2')]
    assert sample('''Fixed Lab 1, compiled with the gcc compiler that comes with OS X on my local machine''') == [('lab', '1')]
    assert sample('''git submit partial lab8: all up to B:6-complete''') == [('lab', '8')]
    assert sample('''had extension, submitting lab5 complete''') == [('lab', '5')]
    assert sample('''homework 1''') == [('hw', '1')]
    assert sample('''homework 10 complete''') == [('hw', '10')]
    assert sample('''homework 11 complete''') == [('hw', '11')]
    assert sample('''homework 12 complete''') == [('hw', '12')]
    assert sample('''Homework 12 complete''') == [('hw', '12')]
    assert sample('''Homework 13 complete''') == [('hw', '13')]
    assert sample('''homework 13 complete''') == [('hw', '13')]
    assert sample('''Homework 14 to section B''') == [('hw', '14')]
    assert sample('''homework 8''') == [('hw', '8')]
    assert sample('''homework 8 complete''') == [('hw', '8')]
    assert sample('''homework1 finished''') == [('hw', '1')]
    assert sample('''homework11 complete''') == [('hw', '11')]
    assert sample('''homework9 complete''') == [('hw', '9')]
    assert sample('''hw 11 complete''') == [('hw', '11')]
    assert sample('''hw 15 complete''') == [('hw', '15')]
    assert sample('''hw 17 exercise files''') == [('hw', '17')]
    assert sample('''hw 19 complete''') == [('hw', '19')]
    assert sample('''hw 19 part C complete''') == [('hw', '19')]
    assert sample('''hw1 complete''') == [('hw', '1')]
    assert sample('''hw1 submit completed''') == [('hw', '1')]
    assert sample('''hw10 complete''') == [('hw', '10')]
    assert sample('''hw14 complete''') == [('hw', '14')]
    assert sample('''hw15 complete''') == [('hw', '15')]
    assert sample('''hw15 examples''') == [('hw', '15')]
    assert sample('''hw16 complete''') == [('hw', '16')]
    assert sample('''hw17 complete''') == [('hw', '17')]
    assert sample('''HW18 complete''') == [('hw', '18')]
    assert sample('''hw18 complete''') == [('hw', '18')]
    assert sample('''hw18 with all constructors''') == [('hw', '18')]
    assert sample('''hw18 with makecopy''') == [('hw', '18')]
    assert sample('''hw18, now with get methods''') == [('hw', '18')]
    assert sample('''HW19 C: final commit post-increment''') == [('hw', '19')]
    assert sample('''HW19 C: pre-increment operator + Makefile''') == [('hw', '19')]
    assert sample('''hw19 complete''') == [('hw', '19')]
    assert sample('''hw19 part C''') == [('hw', '19')]
    assert sample('''hw19, part C complete''') == [('hw', '19')]
    assert sample('''hw20 complete''') == [('hw', '20')]
    assert sample('''hw21 - late, I know''') == [('hw', '21')]
    assert sample('''hw21 almost complete, problem with indexing''') == [('hw', '21')]
    assert sample('''hw21 complete''') == [('hw', '21')]
    assert sample('''hw22 complete''') == [('hw', '22')]
    assert sample('''hw23 complete''') == [('hw', '23')]
    assert sample('''hw23 submit almost complete: a couple bugs''') == [('hw', '23')]
    assert sample('''hw24''') == [('hw', '24')]
    assert sample('''hw24 fixed''') == [('hw', '24')]
    assert sample('''hw6 final submit''') == [('hw', '6')]
    assert sample('''hw7 complpete''') == [('hw', '7')]
    assert sample('''lab 10 complete''') == [('lab', '10')]
    assert sample('''lab 11''') == [('lab', '11')]
    assert sample('''lab 11 complete''') == [('lab', '11')]
    assert sample('''lab 12 complete''') == [('lab', '12')]
    assert sample('''lab 12 part A complete''') == [('lab', '12')]
    assert sample('''Lab 12 part A complete''') == [('lab', '12')]
    assert sample('''lab 12 part A, added ACK mesages''') == [('lab', '12')]
    assert sample('''lab 12 part A, added ACK messages''') == [('lab', '12')]
    assert sample('''lab 12 part A, added ACK messages updated''') == [('lab', '12')]
    assert sample('''lab 3''') == [('lab', '3')]
    assert sample('''Lab 3 Part C''') == [('lab', '3')]
    assert sample('''Lab 3 resubmit attempt''') == [('lab', '3')]
    assert sample('''Lab 3 with octave1.cpp''') == [('lab', '3')]
    assert sample('''Lab 3 with octaveup1.cpp''') == [('lab', '3')]
    assert sample('''Lab 4 complete''') == [('lab', '4')]
    assert sample('''lab 4 part b''') == [('lab', '4')]
    assert sample('''lab 4 submit''') == [('lab', '4')]
    assert sample('''Lab 4 through part B''') == [('lab', '4')]
    assert sample('''Lab 5 complete''') == [('lab', '5')]
    assert sample('''Lab 6 complete''') == [('lab', '6')]
    assert sample('''Lab 6 part A''') == [('lab', '6')]
    assert sample('''Lab 6 Part A''') == [('lab', '6')]
    assert sample('''Lab 6 part B''') == [('lab', '6')]
    assert sample('''Lab 6 Part B''') == [('lab', '6')]
    assert sample('''Lab 6 part B, start of C''') == [('lab', '6')]
    assert sample('''Lab 6 part C''') == [('lab', '6')]
    assert sample('''Lab 6 Part C''') == [('lab', '6')]
    assert sample('''Lab 6 part D''') == [('lab', '6')]
    assert sample('''Lab 6 Part D''') == [('lab', '6')]
    assert sample('''Lab 6 part E''') == [('lab', '6')]
    assert sample('''Lab 6 part F''') == [('lab', '6')]
    assert sample('''Lab 6 Part F''') == [('lab', '6')]
    assert sample('''Lab 6 parts A,B''') == [('lab', '6')]
    assert sample('''lab 7 complete''') == [('lab', '7')]
    assert sample('''Lab 7 part B: Makefile''') == [('lab', '7')]
    assert sample('''lab 7 part b: makefile''') == [('lab', '7')]
    assert sample('''Lab 7 part C''') == [('lab', '7')]
    assert sample('''Lab 7 Part C, fixed proto-game''') == [('lab', '7')]
    assert sample('''Lab 7 part C: proto-game''') == [('lab', '7')]
    assert sample('''lab 7 part c: proto-game''') == [('lab', '7')]
    assert sample('''Lab 7 part D: DVDArray''') == [('lab', '7')]
    assert sample('''Lab 7 pert B: Makefile''') == [('lab', '7')]
    assert sample('''Lab 8 part B: classes shape rectangle circle''') == [('lab', '8')]
    assert sample('''Lab 8 Part C, early proto-draw''') == [('lab', '8')]
    assert sample('''Lab 8 part C: early version of proto-draw''') == [('lab', '8')]
    assert sample('''Lab 8: committing Part A, some bugs still present''') == [('lab', '8')]
    assert sample('''Lab 9 complete''') == [('lab', '9')]
    assert sample('''lab 9 part a complete''') == [('lab', '9')]
    assert sample('''lab 9 part c''') == [('lab', '9')]
    assert sample('''Lab 9 up to D11''') == [('lab', '9')]
    assert sample('''lab12 done''') == [('lab', '12')]
    assert sample('''lab12 part a done''') == [('lab', '12')]
    assert sample('''Lab12 part B''') == [('lab', '12')]
    assert sample('''lab4 Part A complete''') == [('lab', '4')]
    assert sample('''lab4 Part B complete''') == [('lab', '4')]
    assert sample('''lab5 up until c''') == [('lab', '5')]
    assert sample('''lab6 partway through step E''') == [('lab', '6')]
    assert sample('''Lab7 part B: Makefile''') == [('lab', '7')]
    assert sample('''Lab7, Part D, In Progress''') == [('lab', '7')]
    assert sample('''lab8 complete''') == [('lab', '8')]
    assert sample('''lab9 part b''') == [('lab', '9')]
    assert sample('''lab9 part B, pre-multiple objects''') == [('lab', '9')]
    assert sample('''more lab 12 part A DONE''') == [('lab', '12')]
    assert sample('''more of lab 8''') == [('lab', '8')]
    assert sample('''put lab 12 in the right folder''') == [('lab', '12')]
    assert sample('''re-submit hw23 complete''') == [('hw', '23')]
    assert sample('''rename hw3/arithmetic''') == [('hw', '3')]
    assert sample('''rename hw3/casting to miscOperators''') == [('hw', '3')]
    assert sample('''rename hw8/stddev2''') == [('hw', '8')]
    assert sample('''resubmitting because of hw10 error''') == [('hw', '10')]
    assert sample('''retry lab3 subission''') == [('lab', '3')]
    assert sample('''saving lab 2 part A''') == [('lab', '2')]
    assert sample('''saving lab 2 part B''') == [('lab', '2')]
    assert sample('''saving lab2 part b''') == [('lab', '2')]
    assert sample('''submit hw1 complete''') == [('hw', '1')]
    assert sample('''Submit hw10 complete''') == [('hw', '10')]
    assert sample('''submit hw10 complete''') == [('hw', '10')]
    assert sample('''submit hw11 complete''') == [('hw', '11')]
    assert sample('''Submit hw11 complete''') == [('hw', '11')]
    assert sample('''submit hw12 complete''') == [('hw', '12')]
    assert sample('''submit HW12 complete''') == [('hw', '12')]
    assert sample('''submit hw13 complete''') == [('hw', '13')]
    assert sample('''submit hw13 complete with doxygen''') == [('hw', '13')]
    assert sample('''submit hw14''') == [('hw', '14')]
    assert sample('''submit hw14 complete''') == [('hw', '14')]
    assert sample('''submit hw15 complete''') == [('hw', '15')]
    assert sample('''submit hw16 complete''') == [('hw', '16')]
    assert sample('''submit hw17 complete''') == [('hw', '17')]
    assert sample('''submit HW18 complete''') == [('hw', '18')]
    assert sample('''submit hw18 complete''') == [('hw', '18')]
    assert sample('''submit hw18 correct complete''') == [('hw', '18')]
    assert sample('''submit hw19''') == [('hw', '19')]
    assert sample('''submit hw19 complete''') == [('hw', '19')]
    assert sample('''submit hw19 part C''') == [('hw', '19')]
    assert sample('''submit hw2 complete''') == [('hw', '2')]
    assert sample('''submit hw20 can't get assignment operator to work properly ( operator= )''') == [('hw', '20')]
    assert sample('''submit hw20 complete''') == [('hw', '20')]
    assert sample('''submit hw21''') == [('hw', '21')]
    assert sample('''submit hw21 complete''') == [('hw', '21')]
    assert sample('''submit hw21 complete: can't compile- what am I doing wrong?''') == [('hw', '21')]
    assert sample('''submit hw21 correction complete''') == [('hw', '21')]
    assert sample('''submit HW22 complete''') == [('hw', '22')]
    assert sample('''submit hw22 complete''') == [('hw', '22')]
    assert sample('''submit hw23''') == [('hw', '23')]
    assert sample('''submit hw23 complete''') == [('hw', '23')]
    assert sample('''submit hw24''') == [('hw', '24')]
    assert sample('''submit hw24 complete''') == [('hw', '24')]
    assert sample('''submit hw24 complete- a few bugs still present''') == [('hw', '24')]
    assert sample('''submit hw25''') == [('hw', '25')]
    assert sample('''submit hw25 complete''') == [('hw', '25')]
    assert sample('''submit hw3''') == [('hw', '3')]
    assert sample('''submit hw3 complete''') == [('hw', '3')]
    assert sample('''submit hw4 complete''') == [('hw', '4')]
    assert sample('''submit hw5 complete''') == [('hw', '5')]
    assert sample('''submit hw6 complete''') == [('hw', '6')]
    assert sample('''submit hw7 complete''') == [('hw', '7')]
    assert sample('''submit hw8 complete''') == [('hw', '8')]
    assert sample('''submit hw8, save lab3''') == [('hw', '8'), ('lab', '3')]
    assert sample('''submit hw9 complete''') == [('hw', '9')]
    assert sample('''submit Lab 10 complete''') == [('lab', '10')]
    assert sample('''submit Lab 11 complete''') == [('lab', '11')]
    assert sample('''submit lab 12 complete''') == [('lab', '12')]
    assert sample('''submit lab 7 complete''') == [('lab', '7')]
    assert sample('''submit lab 8 complete''') == [('lab', '8')]
    assert sample('''submit lab10 complete''') == [('lab', '10')]
    assert sample('''submit lab11 complete''') == [('lab', '11')]
    assert sample('''submit Lab11 complete''') == [('lab', '11')]
    assert sample('''submit lab12 complete''') == [('lab', '12')]
    assert sample('''submit Lab12 complete''') == [('lab', '12')]
    assert sample('''submit lab2 A-E''') == [('lab', '2')]
    assert sample('''submit lab2 cmplete''') == [('lab', '2')]
    assert sample('''submit lab2 complete''') == [('lab', '2')]
    assert sample('''submit lab3''') == [('lab', '3')]
    assert sample('''submit lab4 all parts except C6. complete''') == [('lab', '4')]
    assert sample('''submit lab4 complete''') == [('lab', '4')]
    assert sample('''submit lab4 through part b''') == [('lab', '4')]
    assert sample('''submit lab5 complete''') == [('lab', '5')]
    assert sample('''submit lab6 A-D''') == [('lab', '6')]
    assert sample('''submit lab6 A-E complete, F:1-2 complete''') == [('lab', '6')]
    assert sample('''submit lab6 complete''') == [('lab', '6')]
    assert sample('''submit lab7 complete''') == [('lab', '7')]
    assert sample('''submit lab8 complete''') == [('lab', '8')]
    assert sample('''submit lab9 complete''') == [('lab', '9')]
    assert sample('''submit lab9 part B''') == [('lab', '9')]
    assert sample('''submit lab9 part c''') == [('lab', '9')]
    assert sample('''submit lab9 part D complete''') == [('lab', '9')]
    assert sample('''sumbit lab2 complete''') == [('lab', '2')]
    assert sample('''sumbmit lab 2''') == [('lab', '2')]
    assert sample('''this is hw 1''') == [('hw', '1')]
    assert sample('''This is my first attempt to submit lab 2''') == [('lab', '2')]
    assert sample('''this is my first attempt to turn in homework 3''') == [('hw', '3')]
    assert sample('''This is my first attempt to turn in homework 6''') == [('hw', '6')]
    assert sample('''this is my first attempt to upload homework 1''') == [('hw', '1')]
    assert sample('''This is my first attempt to upload homework 5''') == [('hw', '5')]
    assert sample('''This is my first attempt to upload homework 7because originally my homework files were stored in the SD directory, so I moved them to hw6, and manually deleted the files on stogit through the web interface''') == [('hw', '7'), ('hw', '6')]
    assert sample('''This is my revision to homework 7''') == [('hw', '7')]
    assert sample('''this is my second attempt at submittiing lab 2''') == [('lab', '2')]
    assert sample('''this is my second attempt to upload homework 1''') == [('hw', '1')]
    assert sample('''This is my second attempt to upload homework 6''') == [('hw', '6')]



from pyfakefs import fake_filesystem
from stograde.common.dirsize import dirsize


def test_chdir(fs):
    os_module = fake_filesystem.FakeOsModule(fs)

    fs.create_dir('/folder')
    fs.create_file('/folder/file1', contents='0123456789')
    fs.create_file('/folder/file2', contents='0123456789')
    fs.create_file('/folder/file3', contents='0123456789')
    fs.create_file('/folder/file4', contents='0123456789')
    fs.create_file('/folder/file5', contents='0123456789')

    print(dirsize('/folder'))



from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Dict, List, TYPE_CHECKING

if TYPE_CHECKING:
    from ..process_assignment.assignment_status import AssignmentStatus
    from ..process_assignment.record_result import RecordResult


@dataclass
class StudentResult:
    name: str
    results: List['RecordResult'] = field(default_factory=list)
    homeworks: Dict[str, 'AssignmentStatus'] = field(default_factory=OrderedDict)
    labs: Dict[str, 'AssignmentStatus'] = field(default_factory=OrderedDict)
    worksheets: Dict[str, 'AssignmentStatus'] = field(default_factory=OrderedDict)
    unmerged_branches: List[str] = field(default_factory=list)
    error: str = ''

    def assignments(self) -> Dict[str, 'AssignmentStatus']:
        assignments = {}
        assignments.update(self.homeworks)
        assignments.update(self.labs)
        assignments.update(self.worksheets)
        return assignments



from dataclasses import dataclass, field
from typing import List


@dataclass
class SubmissionWarnings:
    """Track any warnings about the assignment"""
    assignment_missing: bool = False  # No assignment submission could be found
    recording_err: str = None  # Something raised an error during recording
    unmerged_branches: List[str] = field(default_factory=list)  # There are unmerged branches that might have more code
